#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""This module contains a Google Cloud Vertex AI hook."""

from __future__ import annotations

from collections.abc import Sequence
from typing import TYPE_CHECKING

from google.api_core.client_options import ClientOptions
from google.api_core.gapic_v1.method import DEFAULT, _MethodDefault
from google.cloud.aiplatform_v1 import EndpointServiceClient

from airflow.providers.google.common.consts import CLIENT_INFO
from airflow.providers.google.common.hooks.base_google import GoogleBaseHook
from airflow.providers.google.common.hooks.operation_helpers import OperationHelper

if TYPE_CHECKING:
    from google.api_core.operation import Operation
    from google.api_core.retry import Retry
    from google.cloud.aiplatform_v1.services.endpoint_service.pagers import ListEndpointsPager
    from google.cloud.aiplatform_v1.types import DeployedModel, Endpoint
    from google.protobuf.field_mask_pb2 import FieldMask


class EndpointServiceHook(GoogleBaseHook, OperationHelper):
    """Hook for Google Cloud Vertex AI Endpoint Service APIs."""

    def get_endpoint_service_client(self, region: str | None = None) -> EndpointServiceClient:
        """Return EndpointServiceClient."""
        if region and region != "global":
            client_options = ClientOptions(api_endpoint=f"{region}-aiplatform.googleapis.com:443")
        else:
            client_options = ClientOptions()

        return EndpointServiceClient(
            credentials=self.get_credentials(), client_info=CLIENT_INFO, client_options=client_options
        )

    @staticmethod
    def extract_endpoint_id(obj: dict) -> str:
        """Return unique id of the endpoint."""
        return obj["name"].rpartition("/")[-1]

    @staticmethod
    def extract_deployed_model_id(obj: dict) -> str:
        """Return unique id of the deploy model."""
        return obj["deployed_model"]["id"]

    @GoogleBaseHook.fallback_to_default_project_id
    def create_endpoint(
        self,
        project_id: str,
        region: str,
        endpoint: Endpoint | dict,
        endpoint_id: str | None = None,
        retry: Retry | _MethodDefault = DEFAULT,
        timeout: float | None = None,
        metadata: Sequence[tuple[str, str]] = (),
    ) -> Operation:
        """
        Create an Endpoint.

        :param project_id: Required. The ID of the Google Cloud project that the service belongs to.
        :param region: Required. The ID of the Google Cloud region that the service belongs to.
        :param endpoint: Required. The Endpoint to create.
        :param endpoint_id: The ID of Endpoint. This value should be 1-10 characters, and valid characters
            are /[0-9]/. If not provided, Vertex AI will generate a value for this ID.
        :param retry: Designation of what errors, if any, should be retried.
        :param timeout: The timeout for this request.
        :param metadata: Strings which should be sent along with the request as metadata.
        """
        client = self.get_endpoint_service_client(region)
        parent = client.common_location_path(project_id, region)

        result = client.create_endpoint(
            request={
                "parent": parent,
                "endpoint": endpoint,
                "endpoint_id": endpoint_id,
            },
            retry=retry,
            timeout=timeout,
            metadata=metadata,
        )
        return result

    @GoogleBaseHook.fallback_to_default_project_id
    def delete_endpoint(
        self,
        project_id: str,
        region: str,
        endpoint: str,
        retry: Retry | _MethodDefault = DEFAULT,
        timeout: float | None = None,
        metadata: Sequence[tuple[str, str]] = (),
    ) -> Operation:
        """
        Delete an Endpoint.

        :param project_id: Required. The ID of the Google Cloud project that the service belongs to.
        :param region: Required. The ID of the Google Cloud region that the service belongs to.
        :param endpoint: Required. The Endpoint to delete.
        :param retry: Designation of what errors, if any, should be retried.
        :param timeout: The timeout for this request.
        :param metadata: Strings which should be sent along with the request as metadata.
        """
        client = self.get_endpoint_service_client(region)
        name = client.endpoint_path(project_id, region, endpoint)

        result = client.delete_endpoint(
            request={
                "name": name,
            },
            retry=retry,
            timeout=timeout,
            metadata=metadata,
        )
        return result

    @GoogleBaseHook.fallback_to_default_project_id
    def deploy_model(
        self,
        project_id: str,
        region: str,
        endpoint: str,
        deployed_model: DeployedModel | dict,
        traffic_split: Sequence | dict | None = None,
        retry: Retry | _MethodDefault = DEFAULT,
        timeout: float | None = None,
        metadata: Sequence[tuple[str, str]] = (),
    ) -> Operation:
        """
        Deploys a Model into this Endpoint, creating a DeployedModel within it.

        :param project_id: Required. The ID of the Google Cloud project that the service belongs to.
        :param region: Required. The ID of the Google Cloud region that the service belongs to.
        :param endpoint:  Required. The name of the Endpoint resource into which to deploy a Model. Format:
            ``projects/{project}/locations/{location}/endpoints/{endpoint}``
        :param deployed_model:  Required. The DeployedModel to be created within the Endpoint. Note that
            [Endpoint.traffic_split][google.cloud.aiplatform.v1.Endpoint.traffic_split] must be updated for
            the DeployedModel to start receiving traffic, either as part of this call, or via
            [EndpointService.UpdateEndpoint][google.cloud.aiplatform.v1.EndpointService.UpdateEndpoint].
        :param traffic_split:  A map from a DeployedModel's ID to the percentage of this Endpoint's traffic
            that should be forwarded to that DeployedModel.

            If this field is non-empty, then the Endpoint's
            [traffic_split][google.cloud.aiplatform.v1.Endpoint.traffic_split] will be overwritten with it. To
            refer to the ID of the just being deployed Model, a "0" should be used, and the actual ID of the
            new DeployedModel will be filled in its place by this method. The traffic percentage values must
            add up to 100.

            If this field is empty, then the Endpoint's
            [traffic_split][google.cloud.aiplatform.v1.Endpoint.traffic_split] is not updated.
        :param retry: Designation of what errors, if any, should be retried.
        :param timeout: The timeout for this request.
        :param metadata: Strings which should be sent along with the request as metadata.
        """
        client = self.get_endpoint_service_client(region)
        endpoint_path = client.endpoint_path(project_id, region, endpoint)

        result = client.deploy_model(
            request={
                "endpoint": endpoint_path,
                "deployed_model": deployed_model,
                "traffic_split": traffic_split,
            },
            retry=retry,
            timeout=timeout,
            metadata=metadata,
        )
        return result

    @GoogleBaseHook.fallback_to_default_project_id
    def get_endpoint(
        self,
        project_id: str,
        region: str,
        endpoint: str,
        retry: Retry | _MethodDefault = DEFAULT,
        timeout: float | None = None,
        metadata: Sequence[tuple[str, str]] = (),
    ) -> Endpoint:
        """
        Get an Endpoint.

        :param project_id: Required. The ID of the Google Cloud project that the service belongs to.
        :param region: Required. The ID of the Google Cloud region that the service belongs to.
        :param endpoint: Required. The Endpoint to get.
        :param retry: Designation of what errors, if any, should be retried.
        :param timeout: The timeout for this request.
        :param metadata: Strings which should be sent along with the request as metadata.
        """
        client = self.get_endpoint_service_client(region)
        name = client.endpoint_path(project_id, region, endpoint)

        result = client.get_endpoint(
            request={
                "name": name,
            },
            retry=retry,
            timeout=timeout,
            metadata=metadata,
        )
        return result

    @GoogleBaseHook.fallback_to_default_project_id
    def list_endpoints(
        self,
        project_id: str,
        region: str,
        filter: str | None = None,
        page_size: int | None = None,
        page_token: str | None = None,
        read_mask: str | None = None,
        order_by: str | None = None,
        retry: Retry | _MethodDefault = DEFAULT,
        timeout: float | None = None,
        metadata: Sequence[tuple[str, str]] = (),
    ) -> ListEndpointsPager:
        """
        List Endpoints in a Location.

        :param project_id: Required. The ID of the Google Cloud project that the service belongs to.
        :param region: Required. The ID of the Google Cloud region that the service belongs to.
        :param filter: The standard list filter.
            Supported fields:
            -  ``display_name`` supports = and !=.
            -  ``state`` supports = and !=.
            -  ``model_display_name`` supports = and !=
            Some examples of using the filter are:
            -  ``state="JOB_STATE_SUCCEEDED" AND display_name="my_job"``
            -  ``state="JOB_STATE_RUNNING" OR display_name="my_job"``
            -  ``NOT display_name="my_job"``
            -  ``state="JOB_STATE_FAILED"``
        :param page_size: The standard list page size.
        :param page_token: The standard list page token.
        :param read_mask: Mask specifying which fields to read.
        :param order_by: A comma-separated list of fields to order by, sorted in
            ascending order. Use "desc" after a field name for
            descending. Supported fields:

            -  ``display_name``
            -  ``create_time``
            -  ``update_time``

            Example: ``display_name, create_time desc``.
        :param retry: Designation of what errors, if any, should be retried.
        :param timeout: The timeout for this request.
        :param metadata: Strings which should be sent along with the request as metadata.
        """
        client = self.get_endpoint_service_client(region)
        parent = client.common_location_path(project_id, region)

        result = client.list_endpoints(
            request={
                "parent": parent,
                "filter": filter,
                "page_size": page_size,
                "page_token": page_token,
                "read_mask": read_mask,
                "order_by": order_by,
            },
            retry=retry,
            timeout=timeout,
            metadata=metadata,
        )
        return result

    @GoogleBaseHook.fallback_to_default_project_id
    def undeploy_model(
        self,
        project_id: str,
        region: str,
        endpoint: str,
        deployed_model_id: str,
        traffic_split: Sequence | dict | None = None,
        retry: Retry | _MethodDefault = DEFAULT,
        timeout: float | None = None,
        metadata: Sequence[tuple[str, str]] = (),
    ) -> Operation:
        """
        Undeploys a Model from an Endpoint, removing a DeployedModel from it, and freeing all used resources.

        :param project_id: Required. The ID of the Google Cloud project that the service belongs to.
        :param region: Required. The ID of the Google Cloud region that the service belongs to.
        :param endpoint:  Required. The name of the Endpoint resource from which to undeploy a Model.
        :param deployed_model_id:  Required. The ID of the DeployedModel to be undeployed from the Endpoint.
        :param traffic_split:  If this field is provided, then the Endpoint's
            [traffic_split][google.cloud.aiplatform.v1.Endpoint.traffic_split] will be overwritten with it. If
            last DeployedModel is being undeployed from the Endpoint, the [Endpoint.traffic_split] will always
            end up empty when this call returns. A DeployedModel will be successfully undeployed only if it
            doesn't have any traffic assigned to it when this method executes, or if this field unassigns any
            traffic to it.
        :param retry: Designation of what errors, if any, should be retried.
        :param timeout: The timeout for this request.
        :param metadata: Strings which should be sent along with the request as metadata.
        """
        client = self.get_endpoint_service_client(region)
        endpoint_path = client.endpoint_path(project_id, region, endpoint)

        result = client.undeploy_model(
            request={
                "endpoint": endpoint_path,
                "deployed_model_id": deployed_model_id,
                "traffic_split": traffic_split,
            },
            retry=retry,
            timeout=timeout,
            metadata=metadata,
        )
        return result

    @GoogleBaseHook.fallback_to_default_project_id
    def update_endpoint(
        self,
        project_id: str,
        region: str,
        endpoint_id: str,
        endpoint: Endpoint | dict,
        update_mask: FieldMask | dict,
        retry: Retry | _MethodDefault = DEFAULT,
        timeout: float | None = None,
        metadata: Sequence[tuple[str, str]] = (),
    ) -> Endpoint:
        """
        Update an Endpoint.

        :param project_id: Required. The ID of the Google Cloud project that the service belongs to.
        :param region: Required. The ID of the Google Cloud region that the service belongs to.
        :param endpoint:  Required. The Endpoint which replaces the resource on the server.
        :param update_mask:  Required. The update mask applies to the resource. See
            [google.protobuf.FieldMask][google.protobuf.FieldMask].
        :param retry: Designation of what errors, if any, should be retried.
        :param timeout: The timeout for this request.
        :param metadata: Strings which should be sent along with the request as metadata.
        """
        client = self.get_endpoint_service_client(region)
        endpoint["name"] = client.endpoint_path(project_id, region, endpoint_id)

        result = client.update_endpoint(
            request={
                "endpoint": endpoint,
                "update_mask": update_mask,
            },
            retry=retry,
            timeout=timeout,
            metadata=metadata,
        )
        return result



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""This module contains Google Dataplex links."""

from __future__ import annotations

from airflow.providers.google.cloud.links.base import BaseGoogleLink

DATAPLEX_BASE_LINK = "/dataplex/process/tasks"
DATAPLEX_TASK_LINK = DATAPLEX_BASE_LINK + "/{lake_id}.{task_id};location={region}/jobs?project={project_id}"
DATAPLEX_TASKS_LINK = DATAPLEX_BASE_LINK + "?project={project_id}&qLake={lake_id}.{region}"

DATAPLEX_LAKE_LINK = "/dataplex/lakes/{lake_id};location={region}?project={project_id}"
DATAPLEX_CATALOG_ENTRY_GROUPS_LINK = "/dataplex/catalog/entry-groups?project={project_id}"
DATAPLEX_CATALOG_ENTRY_GROUP_LINK = (
    "/dataplex/projects/{project_id}/locations/{location}/entryGroups/{entry_group_id}?project={project_id}"
)
DATAPLEX_CATALOG_ENTRY_TYPE_LINK = (
    "/dataplex/projects/{project_id}/locations/{location}/entryTypes/{entry_type_id}?project={project_id}"
)
DATAPLEX_CATALOG_ENTRY_TYPES_LINK = "/dataplex/catalog/entry-types?project={project_id}"
DATAPLEX_CATALOG_ASPECT_TYPE_LINK = (
    "/dataplex/projects/{project_id}/locations/{location}/aspectTypes/{aspect_type_id}?project={project_id}"
)
DATAPLEX_CATALOG_ASPECT_TYPES_LINK = "/dataplex/catalog/aspect-types?project={project_id}"
DATAPLEX_CATALOG_ENTRY_LINK = "/dataplex/dp-entries/projects/{project_id}/locations/{location}/entryGroups/{entry_group_id}/entries/{entry_id}?project={project_id}"


class DataplexTaskLink(BaseGoogleLink):
    """Helper class for constructing Dataplex Task link."""

    name = "Dataplex Task"
    key = "task_conf"
    format_str = DATAPLEX_TASK_LINK


class DataplexTasksLink(BaseGoogleLink):
    """Helper class for constructing Dataplex Tasks link."""

    name = "Dataplex Tasks"
    key = "tasks_conf"
    format_str = DATAPLEX_TASKS_LINK


class DataplexLakeLink(BaseGoogleLink):
    """Helper class for constructing Dataplex Lake link."""

    name = "Dataplex Lake"
    key = "dataplex_lake_key"
    format_str = DATAPLEX_LAKE_LINK


class DataplexCatalogEntryGroupLink(BaseGoogleLink):
    """Helper class for constructing Dataplex Catalog EntryGroup link."""

    name = "Dataplex Catalog EntryGroup"
    key = "dataplex_catalog_entry_group_key"
    format_str = DATAPLEX_CATALOG_ENTRY_GROUP_LINK


class DataplexCatalogEntryGroupsLink(BaseGoogleLink):
    """Helper class for constructing Dataplex Catalog EntryGroups link."""

    name = "Dataplex Catalog EntryGroups"
    key = "dataplex_catalog_entry_groups_key"
    format_str = DATAPLEX_CATALOG_ENTRY_GROUPS_LINK


class DataplexCatalogEntryTypeLink(BaseGoogleLink):
    """Helper class for constructing Dataplex Catalog EntryType link."""

    name = "Dataplex Catalog EntryType"
    key = "dataplex_catalog_entry_type_key"
    format_str = DATAPLEX_CATALOG_ENTRY_TYPE_LINK


class DataplexCatalogEntryTypesLink(BaseGoogleLink):
    """Helper class for constructing Dataplex Catalog EntryTypes link."""

    name = "Dataplex Catalog EntryTypes"
    key = "dataplex_catalog_entry_types_key"
    format_str = DATAPLEX_CATALOG_ENTRY_TYPES_LINK


class DataplexCatalogAspectTypeLink(BaseGoogleLink):
    """Helper class for constructing Dataplex Catalog AspectType link."""

    name = "Dataplex Catalog AspectType"
    key = "dataplex_catalog_aspect_type_key"
    format_str = DATAPLEX_CATALOG_ASPECT_TYPE_LINK


class DataplexCatalogAspectTypesLink(BaseGoogleLink):
    """Helper class for constructing Dataplex Catalog AspectTypes link."""

    name = "Dataplex Catalog AspectTypes"
    key = "dataplex_catalog_aspect_types_key"
    format_str = DATAPLEX_CATALOG_ASPECT_TYPES_LINK


class DataplexCatalogEntryLink(BaseGoogleLink):
    """Helper class for constructing Dataplex Catalog Entry link."""

    name = "Dataplex Catalog Entry"
    key = "dataplex_catalog_entry_key"
    format_str = DATAPLEX_CATALOG_ENTRY_LINK



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import asyncio
import logging
from unittest import mock

import pytest

from airflow.providers.google.cloud.triggers.datafusion import DataFusionStartPipelineTrigger
from airflow.triggers.base import TriggerEvent

HOOK_STATUS_STR = "airflow.providers.google.cloud.hooks.datafusion.DataFusionAsyncHook.get_pipeline_status"
CLASSPATH = "airflow.providers.google.cloud.triggers.datafusion.DataFusionStartPipelineTrigger"

TASK_ID = "test_task"
LOCATION = "test-location"
INSTANCE_NAME = "airflow-test-instance"
INSTANCE = {"type": "BASIC", "displayName": INSTANCE_NAME}
PROJECT_ID = "test_project_id"
PIPELINE_NAME = "shrubberyPipeline"
PIPELINE = {"test": "pipeline"}
PIPELINE_ID = "test_pipeline_id"
PIPELINE_TYPE = "batch"
INSTANCE_URL = "http://datafusion.instance.com"
NAMESPACE = "TEST_NAMESPACE"
RUNTIME_ARGS = {"arg1": "a", "arg2": "b"}
TEST_POLL_INTERVAL = 4.0
TEST_GCP_PROJECT_ID = "test-project"


@pytest.fixture
def trigger():
    return DataFusionStartPipelineTrigger(
        instance_url=INSTANCE_URL,
        namespace=NAMESPACE,
        pipeline_name=PIPELINE_NAME,
        pipeline_id=PIPELINE_ID,
        pipeline_type=PIPELINE_TYPE,
        poll_interval=TEST_POLL_INTERVAL,
        gcp_conn_id=TEST_GCP_PROJECT_ID,
    )


class TestDataFusionStartPipelineTrigger:
    def test_start_pipeline_trigger_serialization_should_execute_successfully(self, trigger):
        """
        Asserts that the DataFusionStartPipelineTrigger correctly serializes its arguments
        and classpath.
        """
        classpath, kwargs = trigger.serialize()
        assert classpath == CLASSPATH
        assert kwargs == {
            "instance_url": INSTANCE_URL,
            "namespace": NAMESPACE,
            "pipeline_name": PIPELINE_NAME,
            "pipeline_id": PIPELINE_ID,
            "pipeline_type": PIPELINE_TYPE,
            "gcp_conn_id": TEST_GCP_PROJECT_ID,
            "success_states": None,
        }

    @pytest.mark.asyncio
    @mock.patch(HOOK_STATUS_STR)
    async def test_start_pipeline_trigger_on_success_should_execute_successfully(
        self, mock_pipeline_status, trigger
    ):
        """
        Tests the DataFusionStartPipelineTrigger only fires once the job execution reaches a successful state.
        """
        mock_pipeline_status.return_value = "success"
        generator = trigger.run()
        actual = await generator.asend(None)
        assert (
            TriggerEvent({"status": "success", "message": "Pipeline is running", "pipeline_id": PIPELINE_ID})
            == actual
        )

    @pytest.mark.asyncio
    @mock.patch(HOOK_STATUS_STR)
    async def test_start_pipeline_trigger_running_should_execute_successfully(
        self, mock_pipeline_status, trigger, caplog
    ):
        """
        Test that DataFusionStartPipelineTrigger does not fire while a job is still running.
        """

        mock_pipeline_status.return_value = "pending"
        caplog.set_level(logging.INFO)
        task = asyncio.create_task(trigger.run().__anext__())
        await asyncio.sleep(0.5)

        # TriggerEvent was not returned
        assert task.done() is False

        assert "Pipeline is not still in running state..." in caplog.text
        assert f"Sleeping for {TEST_POLL_INTERVAL} seconds." in caplog.text

        # Prevents error when task is destroyed while in "pending" state
        asyncio.get_event_loop().stop()

    @pytest.mark.asyncio
    @mock.patch(HOOK_STATUS_STR)
    async def test_start_pipeline_trigger_error_should_execute_successfully(
        self, mock_pipeline_status, trigger
    ):
        """
        Test that DataFusionStartPipelineTrigger fires the correct event in case of an error.
        """
        mock_pipeline_status.return_value = "error"

        generator = trigger.run()
        actual = await generator.asend(None)
        assert TriggerEvent({"status": "error", "message": "error"}) == actual

    @pytest.mark.asyncio
    @mock.patch(HOOK_STATUS_STR)
    async def test_start_pipeline_trigger_exception_should_execute_successfully(
        self, mock_pipeline_status, trigger
    ):
        """
        Test that DataFusionStartPipelineTrigger fires the correct event in case of an error.
        """
        mock_pipeline_status.side_effect = Exception("Test exception")

        generator = trigger.run()
        actual = await generator.asend(None)
        assert TriggerEvent({"status": "error", "message": "Test exception"}) == actual



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""
Example Airflow DAG for Google Cloud Dataform service
"""

from __future__ import annotations

import os
from datetime import datetime

from google.cloud.dataform_v1beta1 import WorkflowInvocation, WorkflowInvocationAction

from airflow.models.dag import DAG
from airflow.providers.google.cloud.operators.bigquery import BigQueryDeleteDatasetOperator
from airflow.providers.google.cloud.operators.dataform import (
    DataformCancelWorkflowInvocationOperator,
    DataformCreateCompilationResultOperator,
    DataformCreateRepositoryOperator,
    DataformCreateWorkflowInvocationOperator,
    DataformCreateWorkspaceOperator,
    DataformDeleteRepositoryOperator,
    DataformDeleteWorkspaceOperator,
    DataformGetCompilationResultOperator,
    DataformGetWorkflowInvocationOperator,
    DataformInstallNpmPackagesOperator,
    DataformMakeDirectoryOperator,
    DataformQueryWorkflowInvocationActionsOperator,
    DataformRemoveDirectoryOperator,
    DataformRemoveFileOperator,
    DataformWriteFileOperator,
)
from airflow.providers.google.cloud.sensors.dataform import (
    DataformWorkflowInvocationActionStateSensor,
    DataformWorkflowInvocationStateSensor,
)
from airflow.providers.google.cloud.utils.dataform import make_initialization_workspace_flow

try:
    from airflow.sdk import TriggerRule
except ImportError:
    # Compatibility for Airflow < 3.1
    from airflow.utils.trigger_rule import TriggerRule  # type: ignore[no-redef,attr-defined]

from system.google import DEFAULT_GCP_SYSTEM_TEST_PROJECT_ID

ENV_ID = os.environ.get("SYSTEM_TESTS_ENV_ID", "default")
PROJECT_ID = os.environ.get("SYSTEM_TESTS_GCP_PROJECT") or DEFAULT_GCP_SYSTEM_TEST_PROJECT_ID

DAG_ID = "dataform"

REPOSITORY_ID = f"example_dataform_repository_{ENV_ID}"
REGION = "us-central1"
WORKSPACE_ID = f"example_dataform_workspace_{ENV_ID}"
DATAFORM_SCHEMA_NAME = f"schema_{DAG_ID}_{ENV_ID}"

# This DAG is not self-run we need to do some extra configuration to execute it in automation process
with DAG(
    dag_id=DAG_ID,
    schedule="@once",
    start_date=datetime(2021, 1, 1),
    catchup=False,
    tags=["example", "dataform"],
) as dag:
    # [START howto_operator_create_repository]
    make_repository = DataformCreateRepositoryOperator(
        task_id="make-repository",
        project_id=PROJECT_ID,
        region=REGION,
        repository_id=REPOSITORY_ID,
    )
    # [END howto_operator_create_repository]

    # [START howto_operator_create_workspace]
    make_workspace = DataformCreateWorkspaceOperator(
        task_id="make-workspace",
        project_id=PROJECT_ID,
        region=REGION,
        repository_id=REPOSITORY_ID,
        workspace_id=WORKSPACE_ID,
    )
    # [END howto_operator_create_workspace]

    # [START howto_initialize_workspace]
    first_initialization_step, last_initialization_step = make_initialization_workspace_flow(
        project_id=PROJECT_ID,
        region=REGION,
        repository_id=REPOSITORY_ID,
        workspace_id=WORKSPACE_ID,
        package_name=f"dataform_package_{ENV_ID}",
        without_installation=True,
        dataform_schema_name=DATAFORM_SCHEMA_NAME,
    )
    # [END howto_initialize_workspace]

    # [START howto_operator_install_npm_packages]
    install_npm_packages = DataformInstallNpmPackagesOperator(
        task_id="install-npm-packages",
        project_id=PROJECT_ID,
        region=REGION,
        repository_id=REPOSITORY_ID,
        workspace_id=WORKSPACE_ID,
    )
    # [END howto_operator_install_npm_packages]

    # [START howto_operator_create_compilation_result]
    create_compilation_result = DataformCreateCompilationResultOperator(
        task_id="create-compilation-result",
        project_id=PROJECT_ID,
        region=REGION,
        repository_id=REPOSITORY_ID,
        compilation_result={
            "git_commitish": "main",
            "workspace": (
                f"projects/{PROJECT_ID}/locations/{REGION}/repositories/{REPOSITORY_ID}/"
                f"workspaces/{WORKSPACE_ID}"
            ),
        },
    )
    # [END howto_operator_create_compilation_result]

    # [START howto_operator_get_compilation_result]
    get_compilation_result = DataformGetCompilationResultOperator(
        task_id="get-compilation-result",
        project_id=PROJECT_ID,
        region=REGION,
        repository_id=REPOSITORY_ID,
        compilation_result_id=(
            "{{ task_instance.xcom_pull('create-compilation-result')['name'].split('/')[-1] }}"
        ),
    )
    # [END howto_operator_get_compilation_result]]

    # [START howto_operator_create_workflow_invocation]
    create_workflow_invocation = DataformCreateWorkflowInvocationOperator(
        task_id="create-workflow-invocation",
        project_id=PROJECT_ID,
        region=REGION,
        repository_id=REPOSITORY_ID,
        workflow_invocation={
            "compilation_result": "{{ task_instance.xcom_pull('create-compilation-result')['name'] }}"
        },
    )
    # [END howto_operator_create_workflow_invocation]

    # [START howto_operator_create_workflow_invocation_async]
    create_workflow_invocation_async = DataformCreateWorkflowInvocationOperator(
        task_id="create-workflow-invocation-async",
        project_id=PROJECT_ID,
        region=REGION,
        repository_id=REPOSITORY_ID,
        asynchronous=True,
        workflow_invocation={
            "compilation_result": "{{ task_instance.xcom_pull('create-compilation-result')['name'] }}"
        },
    )

    is_workflow_invocation_done = DataformWorkflowInvocationStateSensor(
        task_id="is-workflow-invocation-done",
        project_id=PROJECT_ID,
        region=REGION,
        repository_id=REPOSITORY_ID,
        workflow_invocation_id=(
            "{{ task_instance.xcom_pull('create-workflow-invocation-async')['name'].split('/')[-1] }}"
        ),
        expected_statuses={WorkflowInvocation.State.SUCCEEDED},
    )
    # [END howto_operator_create_workflow_invocation_async]

    # [START howto_operator_create_workflow_invocation_action_async]
    is_workflow_invocation_action_done = DataformWorkflowInvocationActionStateSensor(
        task_id="is-workflow-invocation-action-done",
        project_id=PROJECT_ID,
        region=REGION,
        repository_id=REPOSITORY_ID,
        workflow_invocation_id=(
            "{{ task_instance.xcom_pull('create-workflow-invocation-async')['name'].split('/')[-1] }}"
        ),
        target_name="first_view",
        expected_statuses={WorkflowInvocationAction.State.SUCCEEDED},
        failure_statuses={
            WorkflowInvocationAction.State.SKIPPED,
            WorkflowInvocationAction.State.DISABLED,
            WorkflowInvocationAction.State.CANCELLED,
            WorkflowInvocationAction.State.FAILED,
        },
    )
    # [END howto_operator_create_workflow_invocation_action_async]

    # [START howto_operator_get_workflow_invocation]
    get_workflow_invocation = DataformGetWorkflowInvocationOperator(
        task_id="get-workflow-invocation",
        project_id=PROJECT_ID,
        region=REGION,
        repository_id=REPOSITORY_ID,
        workflow_invocation_id=(
            "{{ task_instance.xcom_pull('create-workflow-invocation')['name'].split('/')[-1] }}"
        ),
    )
    # [END howto_operator_get_workflow_invocation]

    # [START howto_operator_query_workflow_invocation_actions]
    query_workflow_invocation_actions = DataformQueryWorkflowInvocationActionsOperator(
        task_id="query-workflow-invocation-actions",
        project_id=PROJECT_ID,
        region=REGION,
        repository_id=REPOSITORY_ID,
        workflow_invocation_id=(
            "{{ task_instance.xcom_pull('create-workflow-invocation')['name'].split('/')[-1] }}"
        ),
    )
    # [END howto_operator_query_workflow_invocation_actions]

    create_workflow_invocation_for_cancel = DataformCreateWorkflowInvocationOperator(
        task_id="create-workflow-invocation-for-cancel",
        project_id=PROJECT_ID,
        region=REGION,
        repository_id=REPOSITORY_ID,
        workflow_invocation={
            "compilation_result": "{{ task_instance.xcom_pull('create-compilation-result')['name'] }}"
        },
        asynchronous=True,
    )

    # [START howto_operator_cancel_workflow_invocation]
    cancel_workflow_invocation = DataformCancelWorkflowInvocationOperator(
        task_id="cancel-workflow-invocation",
        project_id=PROJECT_ID,
        region=REGION,
        repository_id=REPOSITORY_ID,
        workflow_invocation_id=(
            "{{ task_instance.xcom_pull('create-workflow-invocation-for-cancel')['name'].split('/')[-1] }}"
        ),
    )
    # [END howto_operator_cancel_workflow_invocation]

    # [START howto_operator_make_directory]
    make_test_directory = DataformMakeDirectoryOperator(
        task_id="make-test-directory",
        project_id=PROJECT_ID,
        region=REGION,
        repository_id=REPOSITORY_ID,
        workspace_id=WORKSPACE_ID,
        directory_path="test",
    )
    # [END howto_operator_make_directory]

    # [START howto_operator_write_file]
    test_file_content = b"""
    test test for test file
    """
    write_test_file = DataformWriteFileOperator(
        task_id="make-test-file",
        project_id=PROJECT_ID,
        region=REGION,
        repository_id=REPOSITORY_ID,
        workspace_id=WORKSPACE_ID,
        filepath="test/test.txt",
        contents=test_file_content,
    )
    # [END howto_operator_write_file]

    # [START howto_operator_remove_file]
    remove_test_file = DataformRemoveFileOperator(
        task_id="remove-test-file",
        project_id=PROJECT_ID,
        region=REGION,
        repository_id=REPOSITORY_ID,
        workspace_id=WORKSPACE_ID,
        filepath="test/test.txt",
    )
    # [END howto_operator_remove_file]

    # [START howto_operator_remove_directory]
    remove_test_directory = DataformRemoveDirectoryOperator(
        task_id="remove-test-directory",
        project_id=PROJECT_ID,
        region=REGION,
        repository_id=REPOSITORY_ID,
        workspace_id=WORKSPACE_ID,
        directory_path="test",
    )
    # [END howto_operator_remove_directory]

    delete_dataset = BigQueryDeleteDatasetOperator(
        task_id="delete-dataset",
        dataset_id=DATAFORM_SCHEMA_NAME,
        delete_contents=True,
        trigger_rule=TriggerRule.ALL_DONE,
    )

    # [START howto_operator_delete_workspace]
    delete_workspace = DataformDeleteWorkspaceOperator(
        task_id="delete-workspace",
        project_id=PROJECT_ID,
        region=REGION,
        repository_id=REPOSITORY_ID,
        workspace_id=WORKSPACE_ID,
        trigger_rule=TriggerRule.ALL_DONE,
    )
    # [END howto_operator_delete_workspace]

    # [START howto_operator_delete_repository]
    delete_repository = DataformDeleteRepositoryOperator(
        task_id="delete-repository",
        project_id=PROJECT_ID,
        region=REGION,
        repository_id=REPOSITORY_ID,
        trigger_rule=TriggerRule.ALL_DONE,
    )
    # [END howto_operator_delete_repository]

    (
        # TEST SETUP
        make_repository
        >> make_workspace
        # TEST BODY
        >> first_initialization_step
    )
    (
        last_initialization_step
        >> make_test_directory
        >> write_test_file
        >> remove_test_file
        >> remove_test_directory
        >> delete_dataset
    )
    (
        last_initialization_step
        >> install_npm_packages
        >> create_compilation_result
        >> [
            get_compilation_result,
            create_workflow_invocation,
            create_workflow_invocation_async,
            create_workflow_invocation_for_cancel,
        ]
    )
    (
        create_workflow_invocation
        >> [get_workflow_invocation, query_workflow_invocation_actions]
        >> delete_dataset
    )
    (
        create_workflow_invocation_async
        >> [is_workflow_invocation_done, is_workflow_invocation_action_done]
        >> delete_dataset
    )
    create_workflow_invocation_for_cancel >> cancel_workflow_invocation >> delete_dataset
    (
        # TEST TEARDOWN
        delete_dataset >> delete_workspace >> delete_repository
    )

    # ### Everything below this line is not part of example ###
    # ### Just for system tests purpose ###
    from tests_common.test_utils.watcher import watcher

    # This test needs watcher in order to properly mark success/failure
    # when "tearDown" task with trigger rule is part of the DAG
    list(dag.tasks) >> watcher()

from tests_common.test_utils.system_tests import get_test_run  # noqa: E402

# Needed to run the example DAG with pytest (see: tests/system/README.md#run_via_pytest)
test_run = get_test_run(dag)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""
Example Airflow DAG for testing interaction between Google Cloud Storage and local file system.
"""

from __future__ import annotations

import os
from datetime import datetime
from pathlib import Path

from airflow.models.dag import DAG
from airflow.providers.google.cloud.operators.gcs import GCSCreateBucketOperator, GCSDeleteBucketOperator
from airflow.providers.google.cloud.transfers.gcs_to_local import GCSToLocalFilesystemOperator
from airflow.providers.google.cloud.transfers.local_to_gcs import LocalFilesystemToGCSOperator

try:
    from airflow.sdk import TriggerRule
except ImportError:
    # Compatibility for Airflow < 3.1
    from airflow.utils.trigger_rule import TriggerRule  # type: ignore[no-redef,attr-defined]

from system.google import DEFAULT_GCP_SYSTEM_TEST_PROJECT_ID
from system.openlineage.operator import OpenLineageTestOperator

ENV_ID = os.environ.get("SYSTEM_TESTS_ENV_ID", "default")
PROJECT_ID = os.environ.get("SYSTEM_TESTS_GCP_PROJECT") or DEFAULT_GCP_SYSTEM_TEST_PROJECT_ID

DAG_ID = "gcs_upload_download"

BUCKET_NAME = f"bucket_{DAG_ID}_{ENV_ID}"
FILE_NAME = "example_upload.txt"
PATH_TO_SAVED_FILE = "example_upload_download.txt"
UPLOAD_FILE_PATH = str(Path(__file__).parent / "resources" / FILE_NAME)


with DAG(
    DAG_ID,
    schedule="@once",
    start_date=datetime(2021, 1, 1),
    catchup=False,
    tags=["gcs", "example"],
) as dag:
    # [START howto_operator_gcs_create_bucket]
    create_bucket = GCSCreateBucketOperator(
        task_id="create_bucket",
        bucket_name=BUCKET_NAME,
        project_id=PROJECT_ID,
    )
    # [END howto_operator_gcs_create_bucket]

    # [START howto_operator_local_filesystem_to_gcs]
    upload_file = LocalFilesystemToGCSOperator(
        task_id="upload_file",
        src=UPLOAD_FILE_PATH,
        dst=FILE_NAME,
        bucket=BUCKET_NAME,
    )
    # [END howto_operator_local_filesystem_to_gcs]

    # [START howto_operator_gcs_download_file_task]
    download_file = GCSToLocalFilesystemOperator(
        task_id="download_file",
        object_name=FILE_NAME,
        bucket=BUCKET_NAME,
        filename=PATH_TO_SAVED_FILE,
    )
    # [END howto_operator_gcs_download_file_task]

    # [START howto_operator_gcs_delete_bucket]
    delete_bucket = GCSDeleteBucketOperator(task_id="delete_bucket", bucket_name=BUCKET_NAME)
    # [END howto_operator_gcs_delete_bucket]
    delete_bucket.trigger_rule = TriggerRule.ALL_DONE

    check_openlineage_events = OpenLineageTestOperator(
        task_id="check_openlineage_events",
        file_path=str(Path(__file__).parent / "resources" / "openlineage" / "gcs_upload_download.json"),
    )

    (
        # TEST SETUP
        create_bucket
        >> upload_file
        # TEST BODY
        >> download_file
        # TEST TEARDOWN
        >> [delete_bucket, check_openlineage_events]
    )

    from tests_common.test_utils.watcher import watcher

    # This test needs watcher in order to properly mark success/failure
    # when "tearDown" task with trigger rule is part of the DAG
    list(dag.tasks) >> watcher()


from tests_common.test_utils.system_tests import get_test_run  # noqa: E402

# Needed to run the example DAG with pytest (see: tests/system/README.md#run_via_pytest)
test_run = get_test_run(dag)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest import mock

import pytest
from google.cloud.workflows.executions_v1beta import Execution

from airflow.exceptions import AirflowException
from airflow.providers.google.cloud.sensors.workflows import WorkflowExecutionSensor

BASE_PATH = "airflow.providers.google.cloud.sensors.workflows.{}"
LOCATION = "europe-west1"
WORKFLOW_ID = "workflow_id"
EXECUTION_ID = "execution_id"
PROJECT_ID = "airflow-testing"
METADATA = None
TIMEOUT = None
RETRY = None
GCP_CONN_ID = "test-conn"
IMPERSONATION_CHAIN = None


class TestWorkflowExecutionSensor:
    @mock.patch(BASE_PATH.format("WorkflowsHook"))
    def test_poke_success(self, mock_hook):
        mock_hook.return_value.get_execution.return_value = mock.MagicMock(state=Execution.State.SUCCEEDED)
        op = WorkflowExecutionSensor(
            task_id="test_task",
            workflow_id=WORKFLOW_ID,
            execution_id=EXECUTION_ID,
            location=LOCATION,
            project_id=PROJECT_ID,
            retry=RETRY,
            request_timeout=TIMEOUT,
            metadata=METADATA,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        result = op.poke({})

        mock_hook.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )

        mock_hook.return_value.get_execution.assert_called_once_with(
            workflow_id=WORKFLOW_ID,
            execution_id=EXECUTION_ID,
            location=LOCATION,
            project_id=PROJECT_ID,
            retry=RETRY,
            timeout=TIMEOUT,
            metadata=METADATA,
        )

        assert result is True

    @mock.patch(BASE_PATH.format("WorkflowsHook"))
    def test_poke_wait(self, mock_hook):
        mock_hook.return_value.get_execution.return_value = mock.MagicMock(state=Execution.State.ACTIVE)
        op = WorkflowExecutionSensor(
            task_id="test_task",
            workflow_id=WORKFLOW_ID,
            execution_id=EXECUTION_ID,
            location=LOCATION,
            project_id=PROJECT_ID,
            retry=RETRY,
            request_timeout=TIMEOUT,
            metadata=METADATA,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        result = op.poke({})

        assert result is False

    @mock.patch(BASE_PATH.format("WorkflowsHook"))
    def test_poke_failure(self, mock_hook):
        mock_hook.return_value.get_execution.return_value = mock.MagicMock(state=Execution.State.FAILED)
        op = WorkflowExecutionSensor(
            task_id="test_task",
            workflow_id=WORKFLOW_ID,
            execution_id=EXECUTION_ID,
            location=LOCATION,
            project_id=PROJECT_ID,
            retry=RETRY,
            request_timeout=TIMEOUT,
            metadata=METADATA,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        with pytest.raises(AirflowException):
            op.poke({})



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from airflow.providers.google.cloud.links.managed_kafka import (
    ApacheKafkaClusterLink,
    ApacheKafkaClusterListLink,
    ApacheKafkaConsumerGroupLink,
    ApacheKafkaTopicLink,
)

TEST_LOCATION = "test-location"
TEST_CLUSTER_ID = "test-cluster-id"
TEST_PROJECT_ID = "test-project-id"
TEST_TOPIC_ID = "test-topic-id"
TEST_CONSUMER_GROUP_ID = "test-consumer-group-id"
EXPECTED_MANAGED_KAFKA_CLUSTER_LINK_NAME = "Apache Kafka Cluster"
EXPECTED_MANAGED_KAFKA_CLUSTER_LINK_KEY = "cluster_conf"
EXPECTED_MANAGED_KAFKA_CLUSTER_LINK_FORMAT_STR = (
    "/managedkafka/{location}/clusters/{cluster_id}?project={project_id}"
)
EXPECTED_MANAGED_KAFKA_CLUSTER_LIST_LINK_NAME = "Apache Kafka Cluster List"
EXPECTED_MANAGED_KAFKA_CLUSTER_LIST_LINK_KEY = "cluster_list_conf"
EXPECTED_MANAGED_KAFKA_CLUSTER_LIST_LINK_FORMAT_STR = "/managedkafka/clusters?project={project_id}"
EXPECTED_MANAGED_KAFKA_TOPIC_LINK_NAME = "Apache Kafka Topic"
EXPECTED_MANAGED_KAFKA_TOPIC_LINK_KEY = "topic_conf"
EXPECTED_MANAGED_KAFKA_TOPIC_LINK_FORMAT_STR = (
    "/managedkafka/{location}/clusters/{cluster_id}/topics/{topic_id}?project={project_id}"
)
EXPECTED_MANAGED_KAFKA_CONSUMER_GROUP_LINK_NAME = "Apache Kafka Consumer Group"
EXPECTED_MANAGED_KAFKA_CONSUMER_GROUP_LINK_KEY = "consumer_group_conf"
EXPECTED_MANAGED_KAFKA_CONSUMER_GROUP_LINK_FORMAT_STR = (
    "/managedkafka/{location}/clusters/{cluster_id}/consumer_groups/{consumer_group_id}?project={project_id}"
)


class TestApacheKafkaClusterLink:
    def test_class_attributes(self):
        assert ApacheKafkaClusterLink.key == EXPECTED_MANAGED_KAFKA_CLUSTER_LINK_KEY
        assert ApacheKafkaClusterLink.name == EXPECTED_MANAGED_KAFKA_CLUSTER_LINK_NAME
        assert ApacheKafkaClusterLink.format_str == EXPECTED_MANAGED_KAFKA_CLUSTER_LINK_FORMAT_STR


class TestApacheKafkaClusterListLink:
    def test_class_attributes(self):
        assert ApacheKafkaClusterListLink.key == EXPECTED_MANAGED_KAFKA_CLUSTER_LIST_LINK_KEY
        assert ApacheKafkaClusterListLink.name == EXPECTED_MANAGED_KAFKA_CLUSTER_LIST_LINK_NAME
        assert ApacheKafkaClusterListLink.format_str == EXPECTED_MANAGED_KAFKA_CLUSTER_LIST_LINK_FORMAT_STR


class TestApacheKafkaTopicLink:
    def test_class_attributes(self):
        assert ApacheKafkaTopicLink.key == EXPECTED_MANAGED_KAFKA_TOPIC_LINK_KEY
        assert ApacheKafkaTopicLink.name == EXPECTED_MANAGED_KAFKA_TOPIC_LINK_NAME
        assert ApacheKafkaTopicLink.format_str == EXPECTED_MANAGED_KAFKA_TOPIC_LINK_FORMAT_STR


class TestApacheKafkaConsumerGroupLink:
    def test_class_attributes(self):
        assert ApacheKafkaConsumerGroupLink.key == EXPECTED_MANAGED_KAFKA_CONSUMER_GROUP_LINK_KEY
        assert ApacheKafkaConsumerGroupLink.name == EXPECTED_MANAGED_KAFKA_CONSUMER_GROUP_LINK_NAME
        assert (
            ApacheKafkaConsumerGroupLink.format_str == EXPECTED_MANAGED_KAFKA_CONSUMER_GROUP_LINK_FORMAT_STR
        )



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""This module contains Google Cloud Storage to SFTP operator."""

from __future__ import annotations

import os
from collections.abc import Sequence
from functools import cached_property
from tempfile import NamedTemporaryFile
from typing import TYPE_CHECKING

from airflow.exceptions import AirflowException
from airflow.providers.google.cloud.hooks.gcs import GCSHook
from airflow.providers.google.version_compat import BaseOperator
from airflow.providers.sftp.hooks.sftp import SFTPHook

WILDCARD = "*"

if TYPE_CHECKING:
    from airflow.providers.common.compat.sdk import Context


class GCSToSFTPOperator(BaseOperator):
    """
    Transfer files from a Google Cloud Storage bucket to SFTP server.

    .. code-block:: python

        with models.DAG(
            "example_gcs_to_sftp",
            start_date=datetime(2020, 6, 19),
            schedule=None,
        ) as dag:
            # downloads file to /tmp/sftp/folder/subfolder/file.txt
            copy_file_from_gcs_to_sftp = GCSToSFTPOperator(
                task_id="file-copy-gsc-to-sftp",
                source_bucket="test-gcs-sftp-bucket-name",
                source_object="folder/subfolder/file.txt",
                destination_path="/tmp/sftp",
            )

            # moves file to /tmp/data.txt
            move_file_from_gcs_to_sftp = GCSToSFTPOperator(
                task_id="file-move-gsc-to-sftp",
                source_bucket="test-gcs-sftp-bucket-name",
                source_object="folder/subfolder/data.txt",
                destination_path="/tmp",
                move_object=True,
                keep_directory_structure=False,
            )

    .. seealso::
        For more information on how to use this operator, take a look at the guide:
        :ref:`howto/operator:GCSToSFTPOperator`

    :param source_bucket: The source Google Cloud Storage bucket where the
         object is. (templated)
    :param source_object: The source name of the object to copy in the Google cloud
        storage bucket. (templated)
        You can use only one wildcard for objects (filenames) within your
        bucket. The wildcard can appear inside the object name or at the
        end of the object name. Appending a wildcard to the bucket name is
        unsupported.
    :param destination_path: The sftp remote path. This is the specified directory path for
        uploading to the SFTP server.
    :param keep_directory_structure: (Optional) When set to False the path of the file
        on the bucket is recreated within path passed in destination_path.
    :param create_intermediate_dirs: (Optional) When set to True the intermediate directories
        in the specified file path will be created.
    :param move_object: When move object is True, the object is moved instead
        of copied to the new location. This is the equivalent of a mv command
        as opposed to a cp command.
    :param gcp_conn_id: (Optional) The connection ID used to connect to Google Cloud.
    :param sftp_conn_id: The sftp connection id. The name or identifier for
        establishing a connection to the SFTP server.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    """

    template_fields: Sequence[str] = (
        "source_bucket",
        "source_object",
        "destination_path",
        "impersonation_chain",
    )
    ui_color = "#f0eee4"

    def __init__(
        self,
        *,
        source_bucket: str,
        source_object: str,
        destination_path: str,
        keep_directory_structure: bool = True,
        create_intermediate_dirs: bool = True,
        move_object: bool = False,
        gcp_conn_id: str = "google_cloud_default",
        sftp_conn_id: str = "ssh_default",
        impersonation_chain: str | Sequence[str] | None = None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)

        self.source_bucket = source_bucket
        self.source_object = source_object
        self.destination_path = destination_path
        self.keep_directory_structure = keep_directory_structure
        self.create_intermediate_dirs = create_intermediate_dirs
        self.move_object = move_object
        self.gcp_conn_id = gcp_conn_id
        self.sftp_conn_id = sftp_conn_id
        self.impersonation_chain = impersonation_chain
        self.sftp_dirs = None

    @cached_property
    def sftp_hook(self):
        return SFTPHook(self.sftp_conn_id)

    def execute(self, context: Context):
        gcs_hook = GCSHook(
            gcp_conn_id=self.gcp_conn_id,
            impersonation_chain=self.impersonation_chain,
        )

        if WILDCARD in self.source_object:
            total_wildcards = self.source_object.count(WILDCARD)
            if total_wildcards > 1:
                raise AirflowException(
                    "Only one wildcard '*' is allowed in source_object parameter. "
                    f"Found {total_wildcards} in {self.source_object}."
                )

            prefix, delimiter = self.source_object.split(WILDCARD, 1)
            prefix_dirname = os.path.dirname(prefix)
            objects = gcs_hook.list(self.source_bucket, prefix=prefix, delimiter=delimiter)
            # TODO: After deprecating delimiter and wildcards in source objects,
            #       remove the previous line and uncomment the following:
            # match_glob = f"**/*{delimiter}" if delimiter else None
            # objects = gcs_hook.list(self.source_bucket, prefix=prefix, match_glob=match_glob)

            for source_object in objects:
                destination_path = self._resolve_destination_path(source_object, prefix=prefix_dirname)
                self._copy_single_object(gcs_hook, self.sftp_hook, source_object, destination_path)

            self.log.info("Done. Uploaded '%d' files to %s", len(objects), self.destination_path)
        else:
            destination_path = self._resolve_destination_path(self.source_object)
            self._copy_single_object(gcs_hook, self.sftp_hook, self.source_object, destination_path)
            self.log.info("Done. Uploaded '%s' file to %s", self.source_object, destination_path)

    def _resolve_destination_path(self, source_object: str, prefix: str | None = None) -> str:
        if not self.keep_directory_structure:
            if prefix:
                source_object = os.path.relpath(source_object, start=prefix)
            else:
                source_object = os.path.basename(source_object)
        return os.path.join(self.destination_path, source_object)

    def _copy_single_object(
        self,
        gcs_hook: GCSHook,
        sftp_hook: SFTPHook,
        source_object: str,
        destination_path: str,
    ) -> None:
        """Copy single object."""
        self.log.info(
            "Executing copy of gs://%s/%s to %s",
            self.source_bucket,
            source_object,
            destination_path,
        )

        dir_path = os.path.dirname(destination_path)

        if self.create_intermediate_dirs:
            sftp_hook.create_directory(dir_path)

        with NamedTemporaryFile("w") as tmp:
            gcs_hook.download(
                bucket_name=self.source_bucket,
                object_name=source_object,
                filename=tmp.name,
            )
            sftp_hook.store_file(destination_path, tmp.name)

        if self.move_object:
            self.log.info("Executing delete of gs://%s/%s", self.source_bucket, source_object)
            gcs_hook.delete(self.source_bucket, source_object)

    def get_openlineage_facets_on_start(self):
        from airflow.providers.common.compat.openlineage.facet import Dataset
        from airflow.providers.google.cloud.openlineage.utils import extract_ds_name_from_gcs_path
        from airflow.providers.openlineage.extractors import OperatorLineage

        source_name = extract_ds_name_from_gcs_path(f"{self.source_object}")
        dest_name = f"{self.destination_path}"
        if self.keep_directory_structure:
            dest_name = os.path.join(dest_name, source_name if source_name != "/" else "")
        elif WILDCARD not in self.source_object:
            dest_name = os.path.join(dest_name, os.path.basename(self.source_object))

        dest_name = dest_name.rstrip("/") if dest_name != "/" else "/"

        return OperatorLineage(
            inputs=[
                Dataset(
                    namespace=f"gs://{self.source_bucket}",
                    name=source_name,
                )
            ],
            outputs=[
                Dataset(
                    namespace=f"file://{self.sftp_hook.remote_host}:{self.sftp_hook.port}",
                    name=dest_name,
                )
            ],
        )



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest import mock

from airflow.providers.google.marketing_platform.hooks.display_video import GoogleDisplayVideo360Hook

from unit.google.cloud.utils.base_gcp_mock import mock_base_gcp_hook_default_project_id

API_VERSION = "v4"
GCP_CONN_ID = "google_cloud_default"


class TestGoogleDisplayVideo360Hook:
    def setup_method(self):
        with mock.patch(
            "airflow.providers.google.common.hooks.base_google.GoogleBaseHook.__init__",
            new=mock_base_gcp_hook_default_project_id,
        ):
            self.hook = GoogleDisplayVideo360Hook(api_version=API_VERSION, gcp_conn_id=GCP_CONN_ID)

    @mock.patch(
        "airflow.providers.google.marketing_platform.hooks.display_video.GoogleDisplayVideo360Hook._authorize"
    )
    @mock.patch("airflow.providers.google.marketing_platform.hooks.display_video.build")
    def test_get_conn_to_display_video(self, mock_build, mock_authorize):
        result = self.hook.get_conn_to_display_video()
        mock_build.assert_called_once_with(
            "displayvideo",
            API_VERSION,
            http=mock_authorize.return_value,
            cache_discovery=False,
        )
        assert mock_build.return_value == result

    @mock.patch(
        "airflow.providers.google.marketing_platform.hooks."
        "display_video.GoogleDisplayVideo360Hook.get_conn_to_display_video"
    )
    def test_create_sdf_download_tasks_called_with_params(self, get_conn_to_display_video):
        body_request = {
            "version": "version",
            "partnerId": "partner_id",
            "advertiserId": "advertiser_id",
            "parentEntityFilter": "parent_entity_filter",
            "idFilter": "id_filter",
            "inventorySourceFilter": "inventory_source_filter",
        }

        self.hook.create_sdf_download_operation(body_request=body_request)

        get_conn_to_display_video.return_value.sdfdownloadtasks.return_value.create.assert_called_once_with(
            body=body_request
        )

    @mock.patch(
        "airflow.providers.google.marketing_platform.hooks."
        "display_video.GoogleDisplayVideo360Hook.get_conn_to_display_video"
    )
    def test_create_sdf_download_tasks_called_once(self, get_conn_to_display_video):
        body_request = {
            "version": "version",
            "partnerId": "partner_id",
            "advertiserId": "advertiser_id",
            "parentEntityFilter": "parent_entity_filter",
            "idFilter": "id_filter",
            "inventorySourceFilter": "inventory_source_filter",
        }

        self.hook.create_sdf_download_operation(body_request=body_request)

        get_conn_to_display_video.return_value.sdfdownloadtasks.return_value.create.assert_called_once()

    @mock.patch(
        "airflow.providers.google.marketing_platform.hooks."
        "display_video.GoogleDisplayVideo360Hook.get_conn_to_display_video"
    )
    def test_create_sdf_download_tasks_return_equal_values(self, get_conn_to_display_video):
        response = ["name"]
        body_request = {
            "version": "version",
            "partnerId": "partner_id",
            "advertiserId": "advertiser_id",
            "parentEntityFilter": "parent_entity_filter",
            "idFilter": "id_filter",
            "inventorySourceFilter": "inventory_source_filter",
        }

        # fmt: off
        get_conn_to_display_video.return_value. \
            sdfdownloadtasks.return_value. \
            create.return_value \
            .execute.return_value = response
        # fmt: on

        result = self.hook.create_sdf_download_operation(body_request=body_request)
        assert response == result

    @mock.patch(
        "airflow.providers.google.marketing_platform.hooks."
        "display_video.GoogleDisplayVideo360Hook.get_conn_to_display_video"
    )
    def test_get_sdf_download_tasks_called_with_params(self, get_conn_to_display_video):
        operation_name = "operation_name"
        self.hook.get_sdf_download_operation(operation_name=operation_name)
        # fmt: off
        get_conn_to_display_video.return_value. \
            sdfdownloadtasks.return_value. \
            operations.return_value. \
            get.assert_called_once_with(name=operation_name)
        # fmt: on

    @mock.patch(
        "airflow.providers.google.marketing_platform.hooks."
        "display_video.GoogleDisplayVideo360Hook.get_conn_to_display_video"
    )
    def test_get_sdf_download_tasks_called_once(self, get_conn_to_display_video):
        operation_name = "name"
        self.hook.get_sdf_download_operation(operation_name=operation_name)
        # fmt: off
        get_conn_to_display_video.return_value. \
            sdfdownloadtasks.return_value. \
            operations.return_value. \
            get.assert_called_once()
        # fmt: on

    @mock.patch(
        "airflow.providers.google.marketing_platform.hooks."
        "display_video.GoogleDisplayVideo360Hook.get_conn_to_display_video"
    )
    def get_sdf_download_tasks_return_equal_values(self, get_conn_to_display_video):
        operation_name = "operation"
        response = "response"

        get_conn_to_display_video.return_value.sdfdownloadtasks.return_value.operations.return_value.get = (
            response
        )

        result = self.hook.get_sdf_download_operation(operation_name=operation_name)

        assert operation_name == result

    @mock.patch(
        "airflow.providers.google.marketing_platform.hooks."
        "display_video.GoogleDisplayVideo360Hook.get_conn_to_display_video"
    )
    def test_download_media_called_once(self, get_conn_to_display_video):
        resource_name = "resource_name"

        self.hook.download_media(resource_name=resource_name)
        get_conn_to_display_video.return_value.media.return_value.download_media.assert_called_once()

    @mock.patch(
        "airflow.providers.google.marketing_platform.hooks."
        "display_video.GoogleDisplayVideo360Hook.get_conn_to_display_video"
    )
    def test_download_media_called_once_with_params(self, get_conn_to_display_video):
        resource_name = "resource_name"

        self.hook.download_media(resource_name=resource_name)
        get_conn_to_display_video.return_value.media.return_value.download_media.assert_called_once_with(
            resourceName=resource_name
        )


class TestGoogleDisplayVideo360v2Hook:
    def setup_method(self):
        with mock.patch(
            "airflow.providers.google.common.hooks.base_google.GoogleBaseHook.__init__",
            new=mock_base_gcp_hook_default_project_id,
        ):
            self.api_version = "v2"
            self.hook = GoogleDisplayVideo360Hook(api_version=self.api_version, gcp_conn_id=GCP_CONN_ID)

    @mock.patch(
        "airflow.providers.google.marketing_platform.hooks.display_video.GoogleDisplayVideo360Hook._authorize"
    )
    @mock.patch("airflow.providers.google.marketing_platform.hooks.display_video.build")
    def test_get_conn_to_display_video(self, mock_build, mock_authorize):
        result = self.hook.get_conn_to_display_video()
        mock_build.assert_called_once_with(
            "displayvideo",
            self.api_version,
            http=mock_authorize.return_value,
            cache_discovery=False,
        )
        assert mock_build.return_value == result



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""
Example Airflow DAG that shows how to use Google Dataprep.

This DAG relies on the following OS environment variables

* SYSTEM_TESTS_DATAPREP_TOKEN - Dataprep API access token.
  For generating it please use instruction
  https://docs.trifacta.com/display/DP/Manage+API+Access+Tokens#:~:text=Enable%20individual%20access-,Generate%20New%20Token,-Via%20UI.
"""

from __future__ import annotations

import json
import logging
import os
from datetime import datetime

from airflow import models

from tests_common.test_utils.version_compat import AIRFLOW_V_3_0_PLUS

if AIRFLOW_V_3_0_PLUS:
    from airflow.sdk import task
else:
    # Airflow 2 path
    from airflow.decorators import task  # type: ignore[attr-defined,no-redef]
from airflow.models.baseoperator import chain
from airflow.providers.google.cloud.hooks.dataprep import GoogleDataprepHook
from airflow.providers.google.cloud.operators.dataprep import (
    DataprepCopyFlowOperator,
    DataprepDeleteFlowOperator,
    DataprepGetJobGroupOperator,
    DataprepGetJobsForJobGroupOperator,
    DataprepRunFlowOperator,
    DataprepRunJobGroupOperator,
)
from airflow.providers.google.cloud.operators.gcs import GCSCreateBucketOperator, GCSDeleteBucketOperator
from airflow.providers.google.cloud.sensors.dataprep import DataprepJobGroupIsFinishedSensor

try:
    from airflow.sdk import TriggerRule
except ImportError:
    # Compatibility for Airflow < 3.1
    from airflow.utils.trigger_rule import TriggerRule  # type: ignore[no-redef,attr-defined]

from system.google import DEFAULT_GCP_SYSTEM_TEST_PROJECT_ID
from system.google.gcp_api_client_helpers import create_airflow_connection, delete_airflow_connection

ENV_ID = os.environ.get("SYSTEM_TESTS_ENV_ID", "default")
DAG_ID = "dataprep"

IS_COMPOSER = bool(os.environ.get("COMPOSER_ENVIRONMENT", ""))

CONNECTION_ID = f"connection_{DAG_ID}_{ENV_ID}".replace("-", "_")
DATAPREP_TOKEN = os.environ.get("SYSTEM_TESTS_DATAPREP_TOKEN", "")
GCP_PROJECT_ID = os.environ.get("SYSTEM_TESTS_GCP_PROJECT") or DEFAULT_GCP_SYSTEM_TEST_PROJECT_ID
GCS_BUCKET_NAME = f"dataprep-bucket-{DAG_ID}-{ENV_ID}"
GCS_BUCKET_PATH = f"gs://{GCS_BUCKET_NAME}/task_results/"

DATASET_URI = "gs://airflow-system-tests-resources/dataprep/dataset-00000.parquet"
DATASET_NAME = f"dataset_{DAG_ID}_{ENV_ID}".replace("-", "_")
DATASET_WRANGLED_NAME = f"wrangled_{DATASET_NAME}"
DATASET_WRANGLED_ID = "{{ task_instance.xcom_pull('create_wrangled_dataset')['id'] }}"

FLOW_ID = "{{ task_instance.xcom_pull('create_flow')['id'] }}"
FLOW_COPY_ID = "{{ task_instance.xcom_pull('copy_flow')['id'] }}"
RECIPE_NAME = DATASET_WRANGLED_NAME
WRITE_SETTINGS = {
    "writesettings": [
        {
            "path": GCS_BUCKET_PATH + f"adhoc_{RECIPE_NAME}.csv",
            "action": "create",
            "format": "csv",
        },
    ],
}

log = logging.getLogger(__name__)

with models.DAG(
    DAG_ID,
    schedule="@once",
    start_date=datetime(2021, 1, 1),  # Override to match your needs
    catchup=False,
    tags=["example", "dataprep"],
    render_template_as_native_obj=True,
) as dag:
    create_bucket_task = GCSCreateBucketOperator(
        task_id="create_bucket",
        bucket_name=GCS_BUCKET_NAME,
        project_id=GCP_PROJECT_ID,
    )

    @task
    def create_connection(connection_id: str) -> None:
        create_airflow_connection(
            connection_id=connection_id,
            connection_conf={
                "description": "Example Dataprep connection",
                "conn_type": "dataprep",
                "extra": json.dumps({"token": DATAPREP_TOKEN}),
            },
            is_composer=IS_COMPOSER,
        )

    create_connection_task = create_connection(connection_id=CONNECTION_ID)

    @task
    def create_imported_dataset():
        hook = GoogleDataprepHook(dataprep_conn_id=CONNECTION_ID)
        response = hook.create_imported_dataset(
            body_request={
                "uri": DATASET_URI,
                "name": DATASET_NAME,
            }
        )
        return response

    create_imported_dataset_task = create_imported_dataset()

    @task
    def create_flow():
        hook = GoogleDataprepHook(dataprep_conn_id=CONNECTION_ID)
        response = hook.create_flow(
            body_request={
                "name": f"test_flow_{DAG_ID}_{ENV_ID}",
                "description": "Test flow",
            }
        )
        return response

    create_flow_task = create_flow()

    @task
    def create_wrangled_dataset(flow, imported_dataset):
        hook = GoogleDataprepHook(dataprep_conn_id=CONNECTION_ID)
        response = hook.create_wrangled_dataset(
            body_request={
                "importedDataset": {"id": imported_dataset["id"]},
                "flow": {"id": flow["id"]},
                "name": DATASET_WRANGLED_NAME,
            }
        )
        return response

    create_wrangled_dataset_task = create_wrangled_dataset(create_flow_task, create_imported_dataset_task)

    @task
    def create_output(wrangled_dataset):
        hook = GoogleDataprepHook(dataprep_conn_id=CONNECTION_ID)
        response = hook.create_output_object(
            body_request={
                "execution": "dataflow",
                "profiler": False,
                "flowNodeId": wrangled_dataset["id"],
            }
        )
        return response

    create_output_task = create_output(create_wrangled_dataset_task)

    @task
    def create_write_settings(output):
        hook = GoogleDataprepHook(dataprep_conn_id=CONNECTION_ID)
        response = hook.create_write_settings(
            body_request={
                "path": GCS_BUCKET_PATH + f"adhoc_{RECIPE_NAME}.csv",
                "action": "create",
                "format": "csv",
                "outputObjectId": output["id"],
            }
        )
        return response

    create_write_settings_task = create_write_settings(create_output_task)

    # [START how_to_dataprep_copy_flow_operator]
    copy_task = DataprepCopyFlowOperator(
        task_id="copy_flow",
        dataprep_conn_id=CONNECTION_ID,
        project_id=GCP_PROJECT_ID,
        flow_id=FLOW_ID,
        name=f"copy_{DATASET_NAME}",
    )
    # [END how_to_dataprep_copy_flow_operator]

    # [START how_to_dataprep_run_job_group_operator]
    run_job_group_task = DataprepRunJobGroupOperator(
        task_id="run_job_group",
        dataprep_conn_id=CONNECTION_ID,
        project_id=GCP_PROJECT_ID,
        body_request={
            "wrangledDataset": {"id": DATASET_WRANGLED_ID},
            "overrides": WRITE_SETTINGS,
        },
    )
    # [END how_to_dataprep_run_job_group_operator]

    # [START how_to_dataprep_dataprep_run_flow_operator]
    run_flow_task = DataprepRunFlowOperator(
        task_id="run_flow",
        dataprep_conn_id=CONNECTION_ID,
        project_id=GCP_PROJECT_ID,
        flow_id=FLOW_COPY_ID,
        body_request={},
    )
    # [END how_to_dataprep_dataprep_run_flow_operator]

    # [START how_to_dataprep_get_job_group_operator]
    get_job_group_task = DataprepGetJobGroupOperator(
        task_id="get_job_group",
        dataprep_conn_id=CONNECTION_ID,
        project_id=GCP_PROJECT_ID,
        job_group_id="{{ task_instance.xcom_pull('run_flow')['data'][0]['id'] }}",
        embed="",
        include_deleted=False,
    )
    # [END how_to_dataprep_get_job_group_operator]

    # [START how_to_dataprep_get_jobs_for_job_group_operator]
    get_jobs_for_job_group_task = DataprepGetJobsForJobGroupOperator(
        task_id="get_jobs_for_job_group",
        dataprep_conn_id=CONNECTION_ID,
        job_group_id="{{ task_instance.xcom_pull('run_flow')['data'][0]['id'] }}",
    )
    # [END how_to_dataprep_get_jobs_for_job_group_operator]

    # [START how_to_dataprep_job_group_finished_sensor]
    check_flow_status_sensor = DataprepJobGroupIsFinishedSensor(
        task_id="check_flow_status",
        dataprep_conn_id=CONNECTION_ID,
        job_group_id="{{ task_instance.xcom_pull('run_flow')['data'][0]['id'] }}",
    )
    # [END how_to_dataprep_job_group_finished_sensor]

    # [START how_to_dataprep_job_group_finished_sensor]
    check_job_group_status_sensor = DataprepJobGroupIsFinishedSensor(
        task_id="check_job_group_status",
        dataprep_conn_id=CONNECTION_ID,
        job_group_id="{{ task_instance.xcom_pull('run_job_group')['id'] }}",
    )
    # [END how_to_dataprep_job_group_finished_sensor]

    # [START how_to_dataprep_delete_flow_operator]
    delete_flow_task = DataprepDeleteFlowOperator(
        task_id="delete_flow",
        dataprep_conn_id=CONNECTION_ID,
        flow_id="{{ task_instance.xcom_pull('copy_flow')['id'] }}",
    )
    # [END how_to_dataprep_delete_flow_operator]
    delete_flow_task.trigger_rule = TriggerRule.ALL_DONE

    delete_flow_task_original = DataprepDeleteFlowOperator(
        task_id="delete_flow_original",
        dataprep_conn_id=CONNECTION_ID,
        flow_id="{{ task_instance.xcom_pull('create_flow')['id'] }}",
        trigger_rule=TriggerRule.ALL_DONE,
    )

    @task(trigger_rule=TriggerRule.ALL_DONE)
    def delete_dataset(dataset):
        hook = GoogleDataprepHook(dataprep_conn_id=CONNECTION_ID)
        hook.delete_imported_dataset(dataset_id=dataset["id"])

    delete_dataset_task = delete_dataset(create_imported_dataset_task)

    delete_bucket_task = GCSDeleteBucketOperator(
        task_id="delete_bucket",
        bucket_name=GCS_BUCKET_NAME,
        trigger_rule=TriggerRule.ALL_DONE,
    )

    @task(task_id="delete_connection")
    def delete_connection(connection_id: str) -> None:
        delete_airflow_connection(connection_id=connection_id, is_composer=IS_COMPOSER)

    delete_connection_task = delete_connection(connection_id=CONNECTION_ID)

    chain(
        # TEST SETUP
        create_bucket_task,
        create_connection_task,
        [create_imported_dataset_task, create_flow_task],
        create_wrangled_dataset_task,
        create_output_task,
        create_write_settings_task,
        # TEST BODY
        copy_task,
        [run_job_group_task, run_flow_task],
        [get_job_group_task, get_jobs_for_job_group_task],
        [check_flow_status_sensor, check_job_group_status_sensor],
        # TEST TEARDOWN
        delete_dataset_task,
        [delete_flow_task, delete_flow_task_original],
        [delete_bucket_task, delete_connection_task],
    )

    from tests_common.test_utils.watcher import watcher

    # This test needs watcher in order to properly mark success/failure
    # when "tearDown" task with trigger rule is part of the DAG
    list(dag.tasks) >> watcher()

from tests_common.test_utils.system_tests import get_test_run  # noqa: E402

# Needed to run the example DAG with pytest (see: tests/system/README.md#run_via_pytest)
test_run = get_test_run(dag)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest import mock

import google.api_core.exceptions
import pytest
from google.cloud.bigtable import enums
from google.cloud.bigtable.column_family import MaxVersionsGCRule
from google.cloud.bigtable.instance import Instance

from airflow.exceptions import AirflowException
from airflow.providers.google.cloud.operators.bigtable import (
    BigtableCreateInstanceOperator,
    BigtableCreateTableOperator,
    BigtableDeleteInstanceOperator,
    BigtableDeleteTableOperator,
    BigtableUpdateClusterOperator,
    BigtableUpdateInstanceOperator,
)

PROJECT_ID = "test_project_id"
INSTANCE_ID = "test-instance-id"
CLUSTER_ID = "test-cluster-id"
CLUSTER_ZONE = "us-central1-f"
REPLICATE_CLUSTERS = [
    {"id": "replica-1", "zone": "us-west1-a"},
    {"id": "replica-2", "zone": "us-central1-f"},
    {"id": "replica-3", "zone": "us-east1-d"},
]
GCP_CONN_ID = "test-gcp-conn-id"
IMPERSONATION_CHAIN = ["ACCOUNT_1", "ACCOUNT_2", "ACCOUNT_3"]
NODES = 5
INSTANCE_DISPLAY_NAME = "test instance"
INSTANCE_TYPE = enums.Instance.Type.PRODUCTION
INSTANCE_LABELS = {"env": "sit"}
TABLE_ID = "test-table-id"
INITIAL_SPLIT_KEYS: list = []
EMPTY_COLUMN_FAMILIES: dict = {}


class TestBigtableInstanceCreate:
    @pytest.mark.parametrize(
        ("missing_attribute", "project_id", "instance_id", "main_cluster_id", "main_cluster_zone"),
        [
            ("instance_id", PROJECT_ID, "", CLUSTER_ID, CLUSTER_ZONE),
            ("main_cluster_id", PROJECT_ID, INSTANCE_ID, "", CLUSTER_ZONE),
            ("main_cluster_zone", PROJECT_ID, INSTANCE_ID, CLUSTER_ID, ""),
        ],
    )
    @mock.patch("airflow.providers.google.cloud.operators.bigtable.BigtableHook")
    def test_empty_attribute(
        self,
        mock_hook,
        missing_attribute,
        project_id,
        instance_id,
        main_cluster_id,
        main_cluster_zone,
    ):
        with pytest.raises(AirflowException) as ctx:
            BigtableCreateInstanceOperator(
                project_id=project_id,
                instance_id=instance_id,
                main_cluster_id=main_cluster_id,
                main_cluster_zone=main_cluster_zone,
                task_id="id",
                gcp_conn_id=GCP_CONN_ID,
            )
        err = ctx.value
        assert str(err) == f"Empty parameter: {missing_attribute}"
        mock_hook.assert_not_called()

    @mock.patch("airflow.providers.google.cloud.operators.bigtable.BigtableHook")
    def test_create_instance_that_exists(self, mock_hook):
        mock_hook.return_value.get_instance.return_value = mock.Mock(Instance)

        op = BigtableCreateInstanceOperator(
            project_id=PROJECT_ID,
            instance_id=INSTANCE_ID,
            main_cluster_id=CLUSTER_ID,
            main_cluster_zone=CLUSTER_ZONE,
            task_id="id",
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        op.execute(context={"ti": mock.MagicMock(), "task": mock.MagicMock()})

        mock_hook.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        mock_hook.return_value.create_instance.assert_not_called()

    @mock.patch("airflow.providers.google.cloud.operators.bigtable.BigtableHook")
    def test_create_instance_that_exists_empty_project_id(self, mock_hook):
        mock_hook.return_value.get_instance.return_value = mock.Mock(Instance)

        op = BigtableCreateInstanceOperator(
            instance_id=INSTANCE_ID,
            main_cluster_id=CLUSTER_ID,
            main_cluster_zone=CLUSTER_ZONE,
            task_id="id",
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        op.execute(context={"ti": mock.MagicMock(), "task": mock.MagicMock()})

        mock_hook.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        mock_hook.return_value.create_instance.assert_not_called()

    @mock.patch("airflow.providers.google.cloud.operators.bigtable.BigtableHook")
    def test_different_error_reraised(self, mock_hook):
        mock_hook.return_value.get_instance.return_value = None
        op = BigtableCreateInstanceOperator(
            project_id=PROJECT_ID,
            instance_id=INSTANCE_ID,
            main_cluster_id=CLUSTER_ID,
            main_cluster_zone=CLUSTER_ZONE,
            task_id="id",
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )

        mock_hook.return_value.create_instance.side_effect = mock.Mock(
            side_effect=google.api_core.exceptions.GoogleAPICallError("error")
        )

        with pytest.raises(google.api_core.exceptions.GoogleAPICallError):
            op.execute(None)

        mock_hook.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        mock_hook.return_value.create_instance.assert_called_once_with(
            cluster_nodes=None,
            cluster_storage_type=None,
            instance_display_name=None,
            instance_id=INSTANCE_ID,
            instance_labels=None,
            instance_type=None,
            main_cluster_id=CLUSTER_ID,
            main_cluster_zone=CLUSTER_ZONE,
            project_id=PROJECT_ID,
            replica_clusters=None,
            timeout=None,
        )

    @mock.patch("airflow.providers.google.cloud.operators.bigtable.BigtableHook")
    def test_create_instance_that_doesnt_exists(self, mock_hook):
        mock_hook.return_value.get_instance.return_value = None
        op = BigtableCreateInstanceOperator(
            project_id=PROJECT_ID,
            instance_id=INSTANCE_ID,
            main_cluster_id=CLUSTER_ID,
            main_cluster_zone=CLUSTER_ZONE,
            task_id="id",
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        op.execute(context={"ti": mock.MagicMock(), "task": mock.MagicMock()})
        mock_hook.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        mock_hook.return_value.create_instance.assert_called_once_with(
            cluster_nodes=None,
            cluster_storage_type=None,
            instance_display_name=None,
            instance_id=INSTANCE_ID,
            instance_labels=None,
            instance_type=None,
            main_cluster_id=CLUSTER_ID,
            main_cluster_zone=CLUSTER_ZONE,
            project_id=PROJECT_ID,
            replica_clusters=None,
            timeout=None,
        )

    @mock.patch("airflow.providers.google.cloud.operators.bigtable.BigtableHook")
    def test_create_instance_with_replicas_that_doesnt_exists(self, mock_hook):
        mock_hook.return_value.get_instance.return_value = None
        op = BigtableCreateInstanceOperator(
            project_id=PROJECT_ID,
            instance_id=INSTANCE_ID,
            main_cluster_id=CLUSTER_ID,
            main_cluster_zone=CLUSTER_ZONE,
            replica_clusters=REPLICATE_CLUSTERS,
            task_id="id",
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        op.execute(context={"ti": mock.MagicMock(), "task": mock.MagicMock()})
        mock_hook.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        mock_hook.return_value.create_instance.assert_called_once_with(
            cluster_nodes=None,
            cluster_storage_type=None,
            instance_display_name=None,
            instance_id=INSTANCE_ID,
            instance_labels=None,
            instance_type=None,
            main_cluster_id=CLUSTER_ID,
            main_cluster_zone=CLUSTER_ZONE,
            project_id=PROJECT_ID,
            replica_clusters=REPLICATE_CLUSTERS,
            timeout=None,
        )


class TestBigtableInstanceUpdate:
    @mock.patch("airflow.providers.google.cloud.operators.bigtable.BigtableHook")
    def test_delete_execute(self, mock_hook):
        op = BigtableUpdateInstanceOperator(
            project_id=PROJECT_ID,
            instance_id=INSTANCE_ID,
            instance_display_name=INSTANCE_DISPLAY_NAME,
            instance_type=INSTANCE_TYPE,
            instance_labels=INSTANCE_LABELS,
            task_id="id",
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        op.execute(context={"ti": mock.MagicMock(), "task": mock.MagicMock()})
        mock_hook.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        mock_hook.return_value.update_instance.assert_called_once_with(
            project_id=PROJECT_ID,
            instance_id=INSTANCE_ID,
            instance_display_name=INSTANCE_DISPLAY_NAME,
            instance_type=INSTANCE_TYPE,
            instance_labels=INSTANCE_LABELS,
            timeout=None,
        )

    @mock.patch("airflow.providers.google.cloud.operators.bigtable.BigtableHook")
    def test_update_execute_empty_project_id(self, mock_hook):
        op = BigtableUpdateInstanceOperator(
            instance_id=INSTANCE_ID,
            instance_display_name=INSTANCE_DISPLAY_NAME,
            instance_type=INSTANCE_TYPE,
            instance_labels=INSTANCE_LABELS,
            task_id="id",
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        op.execute(context={"ti": mock.MagicMock(), "task": mock.MagicMock()})
        mock_hook.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        mock_hook.return_value.update_instance.assert_called_once_with(
            project_id=None,
            instance_id=INSTANCE_ID,
            instance_display_name=INSTANCE_DISPLAY_NAME,
            instance_type=INSTANCE_TYPE,
            instance_labels=INSTANCE_LABELS,
            timeout=None,
        )

    @pytest.mark.parametrize(
        ("missing_attribute", "project_id", "instance_id"),
        [("instance_id", PROJECT_ID, "")],
    )
    @mock.patch("airflow.providers.google.cloud.operators.bigtable.BigtableHook")
    def test_empty_attribute(self, mock_hook, missing_attribute, project_id, instance_id):
        with pytest.raises(AirflowException) as ctx:
            BigtableUpdateInstanceOperator(
                project_id=project_id,
                instance_id=instance_id,
                instance_display_name=INSTANCE_DISPLAY_NAME,
                instance_type=INSTANCE_TYPE,
                instance_labels=INSTANCE_LABELS,
                task_id="id",
            )
        err = ctx.value
        assert str(err) == f"Empty parameter: {missing_attribute}"
        mock_hook.assert_not_called()

    @mock.patch("airflow.providers.google.cloud.operators.bigtable.BigtableHook")
    def test_update_instance_that_doesnt_exists(self, mock_hook):
        mock_hook.return_value.get_instance.return_value = None
        op = BigtableUpdateInstanceOperator(
            project_id=PROJECT_ID,
            instance_id=INSTANCE_ID,
            instance_display_name=INSTANCE_DISPLAY_NAME,
            instance_type=INSTANCE_TYPE,
            instance_labels=INSTANCE_LABELS,
            task_id="id",
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        with pytest.raises(AirflowException) as ctx:
            op.execute(None)

        err = ctx.value
        assert str(err) == f"Dependency: instance '{INSTANCE_ID}' does not exist."

        mock_hook.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        mock_hook.return_value.update_instance.assert_not_called()

    @mock.patch("airflow.providers.google.cloud.operators.bigtable.BigtableHook")
    def test_update_instance_that_doesnt_exists_empty_project_id(self, mock_hook):
        mock_hook.return_value.get_instance.return_value = None
        op = BigtableUpdateInstanceOperator(
            instance_id=INSTANCE_ID,
            instance_display_name=INSTANCE_DISPLAY_NAME,
            instance_type=INSTANCE_TYPE,
            instance_labels=INSTANCE_LABELS,
            task_id="id",
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        with pytest.raises(AirflowException) as ctx:
            op.execute(None)

        err = ctx.value
        assert str(err) == f"Dependency: instance '{INSTANCE_ID}' does not exist."

        mock_hook.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        mock_hook.return_value.update_instance.assert_not_called()

    @mock.patch("airflow.providers.google.cloud.operators.bigtable.BigtableHook")
    def test_different_error_reraised(self, mock_hook):
        op = BigtableUpdateInstanceOperator(
            project_id=PROJECT_ID,
            instance_id=INSTANCE_ID,
            instance_display_name=INSTANCE_DISPLAY_NAME,
            instance_type=INSTANCE_TYPE,
            instance_labels=INSTANCE_LABELS,
            task_id="id",
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        mock_hook.return_value.update_instance.side_effect = mock.Mock(
            side_effect=google.api_core.exceptions.GoogleAPICallError("error")
        )

        with pytest.raises(google.api_core.exceptions.GoogleAPICallError):
            op.execute(None)

        mock_hook.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        mock_hook.return_value.update_instance.assert_called_once_with(
            project_id=PROJECT_ID,
            instance_id=INSTANCE_ID,
            instance_display_name=INSTANCE_DISPLAY_NAME,
            instance_type=INSTANCE_TYPE,
            instance_labels=INSTANCE_LABELS,
            timeout=None,
        )


class TestBigtableClusterUpdate:
    @pytest.mark.parametrize(
        ("missing_attribute", "project_id", "instance_id", "cluster_id", "nodes"),
        [
            ("instance_id", PROJECT_ID, "", CLUSTER_ID, NODES),
            ("cluster_id", PROJECT_ID, INSTANCE_ID, "", NODES),
            ("nodes", PROJECT_ID, INSTANCE_ID, CLUSTER_ID, ""),
        ],
    )
    @mock.patch("airflow.providers.google.cloud.operators.bigtable.BigtableHook")
    def test_empty_attribute(self, mock_hook, missing_attribute, project_id, instance_id, cluster_id, nodes):
        with pytest.raises(AirflowException) as ctx:
            BigtableUpdateClusterOperator(
                project_id=project_id,
                instance_id=instance_id,
                cluster_id=cluster_id,
                nodes=nodes,
                task_id="id",
                gcp_conn_id=GCP_CONN_ID,
            )
        err = ctx.value
        assert str(err) == f"Empty parameter: {missing_attribute}"
        mock_hook.assert_not_called()

    @mock.patch("airflow.providers.google.cloud.operators.bigtable.BigtableHook")
    def test_updating_cluster_but_instance_does_not_exists(self, mock_hook):
        mock_hook.return_value.get_instance.return_value = None
        op = BigtableUpdateClusterOperator(
            project_id=PROJECT_ID,
            instance_id=INSTANCE_ID,
            cluster_id=CLUSTER_ID,
            nodes=NODES,
            task_id="id",
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        with pytest.raises(AirflowException) as ctx:
            op.execute(None)

        err = ctx.value
        assert str(err) == f"Dependency: instance '{INSTANCE_ID}' does not exist."
        mock_hook.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        mock_hook.return_value.update_cluster.assert_not_called()

    @mock.patch("airflow.providers.google.cloud.operators.bigtable.BigtableHook")
    def test_updating_cluster_but_instance_does_not_exists_empty_project_id(self, mock_hook):
        mock_hook.return_value.get_instance.return_value = None
        op = BigtableUpdateClusterOperator(
            instance_id=INSTANCE_ID,
            cluster_id=CLUSTER_ID,
            nodes=NODES,
            task_id="id",
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        with pytest.raises(AirflowException) as ctx:
            op.execute(None)

        err = ctx.value
        assert str(err) == f"Dependency: instance '{INSTANCE_ID}' does not exist."
        mock_hook.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        mock_hook.return_value.update_cluster.assert_not_called()

    @mock.patch("airflow.providers.google.cloud.operators.bigtable.BigtableHook")
    def test_updating_cluster_that_does_not_exists(self, mock_hook):
        instance = mock_hook.return_value.get_instance.return_value = mock.Mock(Instance)
        mock_hook.return_value.update_cluster.side_effect = mock.Mock(
            side_effect=google.api_core.exceptions.NotFound("Cluster not found.")
        )
        op = BigtableUpdateClusterOperator(
            project_id=PROJECT_ID,
            instance_id=INSTANCE_ID,
            cluster_id=CLUSTER_ID,
            nodes=NODES,
            task_id="id",
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        with pytest.raises(AirflowException) as ctx:
            op.execute(None)

        err = ctx.value
        assert str(err) == f"Dependency: cluster '{CLUSTER_ID}' does not exist for instance '{INSTANCE_ID}'."
        mock_hook.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        mock_hook.return_value.update_cluster.assert_called_once_with(
            instance=instance, cluster_id=CLUSTER_ID, nodes=NODES
        )

    @mock.patch("airflow.providers.google.cloud.operators.bigtable.BigtableHook")
    def test_updating_cluster_that_does_not_exists_empty_project_id(self, mock_hook):
        instance = mock_hook.return_value.get_instance.return_value = mock.Mock(Instance)
        mock_hook.return_value.update_cluster.side_effect = mock.Mock(
            side_effect=google.api_core.exceptions.NotFound("Cluster not found.")
        )
        op = BigtableUpdateClusterOperator(
            instance_id=INSTANCE_ID,
            cluster_id=CLUSTER_ID,
            nodes=NODES,
            task_id="id",
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        with pytest.raises(AirflowException) as ctx:
            op.execute(None)

        err = ctx.value
        assert str(err) == f"Dependency: cluster '{CLUSTER_ID}' does not exist for instance '{INSTANCE_ID}'."
        mock_hook.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        mock_hook.return_value.update_cluster.assert_called_once_with(
            instance=instance, cluster_id=CLUSTER_ID, nodes=NODES
        )

    @mock.patch("airflow.providers.google.cloud.operators.bigtable.BigtableHook")
    def test_different_error_reraised(self, mock_hook):
        op = BigtableUpdateClusterOperator(
            project_id=PROJECT_ID,
            instance_id=INSTANCE_ID,
            cluster_id=CLUSTER_ID,
            nodes=NODES,
            task_id="id",
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        instance = mock_hook.return_value.get_instance.return_value = mock.Mock(Instance)
        mock_hook.return_value.update_cluster.side_effect = mock.Mock(
            side_effect=google.api_core.exceptions.GoogleAPICallError("error")
        )

        with pytest.raises(google.api_core.exceptions.GoogleAPICallError):
            op.execute(None)

        mock_hook.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        mock_hook.return_value.update_cluster.assert_called_once_with(
            instance=instance, cluster_id=CLUSTER_ID, nodes=NODES
        )


class TestBigtableInstanceDelete:
    @mock.patch("airflow.providers.google.cloud.operators.bigtable.BigtableHook")
    def test_delete_execute(self, mock_hook):
        op = BigtableDeleteInstanceOperator(
            project_id=PROJECT_ID,
            instance_id=INSTANCE_ID,
            task_id="id",
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        op.execute(None)
        mock_hook.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        mock_hook.return_value.delete_instance.assert_called_once_with(
            project_id=PROJECT_ID, instance_id=INSTANCE_ID
        )

    @mock.patch("airflow.providers.google.cloud.operators.bigtable.BigtableHook")
    def test_delete_execute_empty_project_id(self, mock_hook):
        op = BigtableDeleteInstanceOperator(
            instance_id=INSTANCE_ID,
            task_id="id",
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        op.execute(None)
        mock_hook.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        mock_hook.return_value.delete_instance.assert_called_once_with(
            project_id=None, instance_id=INSTANCE_ID
        )

    @pytest.mark.parametrize(
        ("missing_attribute", "project_id", "instance_id"),
        [("instance_id", PROJECT_ID, "")],
    )
    @mock.patch("airflow.providers.google.cloud.operators.bigtable.BigtableHook")
    def test_empty_attribute(self, mock_hook, missing_attribute, project_id, instance_id):
        with pytest.raises(AirflowException) as ctx:
            BigtableDeleteInstanceOperator(project_id=project_id, instance_id=instance_id, task_id="id")
        err = ctx.value
        assert str(err) == f"Empty parameter: {missing_attribute}"
        mock_hook.assert_not_called()

    @mock.patch("airflow.providers.google.cloud.operators.bigtable.BigtableHook")
    def test_deleting_instance_that_doesnt_exists(self, mock_hook):
        op = BigtableDeleteInstanceOperator(
            project_id=PROJECT_ID,
            instance_id=INSTANCE_ID,
            task_id="id",
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        mock_hook.return_value.delete_instance.side_effect = mock.Mock(
            side_effect=google.api_core.exceptions.NotFound("Instance not found.")
        )
        op.execute(None)
        mock_hook.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        mock_hook.return_value.delete_instance.assert_called_once_with(
            project_id=PROJECT_ID, instance_id=INSTANCE_ID
        )

    @mock.patch("airflow.providers.google.cloud.operators.bigtable.BigtableHook")
    def test_deleting_instance_that_doesnt_exists_empty_project_id(self, mock_hook):
        op = BigtableDeleteInstanceOperator(
            instance_id=INSTANCE_ID,
            task_id="id",
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        mock_hook.return_value.delete_instance.side_effect = mock.Mock(
            side_effect=google.api_core.exceptions.NotFound("Instance not found.")
        )
        op.execute(None)
        mock_hook.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        mock_hook.return_value.delete_instance.assert_called_once_with(
            project_id=None, instance_id=INSTANCE_ID
        )

    @mock.patch("airflow.providers.google.cloud.operators.bigtable.BigtableHook")
    def test_different_error_reraised(self, mock_hook):
        op = BigtableDeleteInstanceOperator(
            project_id=PROJECT_ID,
            instance_id=INSTANCE_ID,
            task_id="id",
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        mock_hook.return_value.delete_instance.side_effect = mock.Mock(
            side_effect=google.api_core.exceptions.GoogleAPICallError("error")
        )

        with pytest.raises(google.api_core.exceptions.GoogleAPICallError):
            op.execute(None)

        mock_hook.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        mock_hook.return_value.delete_instance.assert_called_once_with(
            project_id=PROJECT_ID, instance_id=INSTANCE_ID
        )


class TestBigtableTableDelete:
    @mock.patch("airflow.providers.google.cloud.operators.bigtable.BigtableHook")
    def test_delete_execute(self, mock_hook):
        op = BigtableDeleteTableOperator(
            project_id=PROJECT_ID,
            instance_id=INSTANCE_ID,
            table_id=TABLE_ID,
            task_id="id",
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        op.execute(None)
        mock_hook.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        mock_hook.return_value.delete_table.assert_called_once_with(
            project_id=PROJECT_ID, instance_id=INSTANCE_ID, table_id=TABLE_ID
        )

    @pytest.mark.parametrize(
        ("missing_attribute", "project_id", "instance_id", "table_id"),
        [
            ("instance_id", PROJECT_ID, "", TABLE_ID),
            ("table_id", PROJECT_ID, INSTANCE_ID, ""),
        ],
    )
    @mock.patch("airflow.providers.google.cloud.operators.bigtable.BigtableHook")
    def test_empty_attribute(self, mock_hook, missing_attribute, project_id, instance_id, table_id):
        with pytest.raises(AirflowException) as ctx:
            BigtableDeleteTableOperator(
                project_id=project_id,
                instance_id=instance_id,
                table_id=table_id,
                task_id="id",
                gcp_conn_id=GCP_CONN_ID,
            )
        err = ctx.value
        assert str(err) == f"Empty parameter: {missing_attribute}"
        mock_hook.assert_not_called()

    @mock.patch("airflow.providers.google.cloud.operators.bigtable.BigtableHook")
    def test_deleting_table_that_doesnt_exists(self, mock_hook):
        op = BigtableDeleteTableOperator(
            project_id=PROJECT_ID,
            instance_id=INSTANCE_ID,
            table_id=TABLE_ID,
            task_id="id",
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )

        mock_hook.return_value.delete_table.side_effect = mock.Mock(
            side_effect=google.api_core.exceptions.NotFound("Table not found.")
        )
        op.execute(None)
        mock_hook.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        mock_hook.return_value.delete_table.assert_called_once_with(
            project_id=PROJECT_ID, instance_id=INSTANCE_ID, table_id=TABLE_ID
        )

    @mock.patch("airflow.providers.google.cloud.operators.bigtable.BigtableHook")
    def test_deleting_table_that_doesnt_exists_empty_project_id(self, mock_hook):
        op = BigtableDeleteTableOperator(
            instance_id=INSTANCE_ID,
            table_id=TABLE_ID,
            task_id="id",
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )

        mock_hook.return_value.delete_table.side_effect = mock.Mock(
            side_effect=google.api_core.exceptions.NotFound("Table not found.")
        )
        op.execute(None)
        mock_hook.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        mock_hook.return_value.delete_table.assert_called_once_with(
            project_id=None, instance_id=INSTANCE_ID, table_id=TABLE_ID
        )

    @mock.patch("airflow.providers.google.cloud.operators.bigtable.BigtableHook")
    def test_deleting_table_when_instance_doesnt_exists(self, mock_hook):
        op = BigtableDeleteTableOperator(
            project_id=PROJECT_ID,
            instance_id=INSTANCE_ID,
            table_id=TABLE_ID,
            task_id="id",
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )

        mock_hook.return_value.get_instance.return_value = None
        with pytest.raises(AirflowException) as ctx:
            op.execute(None)
        err = ctx.value
        assert str(err) == f"Dependency: instance '{INSTANCE_ID}' does not exist."
        mock_hook.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        mock_hook.return_value.delete_table.assert_not_called()

    @mock.patch("airflow.providers.google.cloud.operators.bigtable.BigtableHook")
    def test_different_error_reraised(self, mock_hook):
        op = BigtableDeleteTableOperator(
            project_id=PROJECT_ID,
            instance_id=INSTANCE_ID,
            table_id=TABLE_ID,
            task_id="id",
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        mock_hook.return_value.delete_table.side_effect = mock.Mock(
            side_effect=google.api_core.exceptions.GoogleAPICallError("error")
        )

        with pytest.raises(google.api_core.exceptions.GoogleAPICallError):
            op.execute(None)

        mock_hook.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        mock_hook.return_value.delete_table.assert_called_once_with(
            project_id=PROJECT_ID, instance_id=INSTANCE_ID, table_id=TABLE_ID
        )


class TestBigtableTableCreate:
    @mock.patch("airflow.providers.google.cloud.operators.bigtable.BigtableHook")
    def test_create_execute(self, mock_hook):
        op = BigtableCreateTableOperator(
            project_id=PROJECT_ID,
            instance_id=INSTANCE_ID,
            table_id=TABLE_ID,
            initial_split_keys=INITIAL_SPLIT_KEYS,
            column_families=EMPTY_COLUMN_FAMILIES,
            task_id="id",
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        instance = mock_hook.return_value.get_instance.return_value = mock.Mock(Instance)
        op.execute(context={"ti": mock.MagicMock(), "task": mock.MagicMock()})
        mock_hook.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        mock_hook.return_value.create_table.assert_called_once_with(
            instance=instance,
            table_id=TABLE_ID,
            initial_split_keys=INITIAL_SPLIT_KEYS,
            column_families=EMPTY_COLUMN_FAMILIES,
        )

    @pytest.mark.parametrize(
        ("missing_attribute", "project_id", "instance_id", "table_id"),
        [
            ("instance_id", PROJECT_ID, "", TABLE_ID),
            ("table_id", PROJECT_ID, INSTANCE_ID, ""),
        ],
    )
    @mock.patch("airflow.providers.google.cloud.operators.bigtable.BigtableHook")
    def test_empty_attribute(self, mock_hook, missing_attribute, project_id, instance_id, table_id):
        with pytest.raises(AirflowException) as ctx:
            BigtableCreateTableOperator(
                project_id=project_id,
                instance_id=instance_id,
                table_id=table_id,
                task_id="id",
                gcp_conn_id=GCP_CONN_ID,
            )
        err = ctx.value
        assert str(err) == f"Empty parameter: {missing_attribute}"
        mock_hook.assert_not_called()

    @mock.patch("airflow.providers.google.cloud.operators.bigtable.BigtableHook")
    def test_instance_not_exists(self, mock_hook):
        op = BigtableCreateTableOperator(
            project_id=PROJECT_ID,
            instance_id=INSTANCE_ID,
            table_id=TABLE_ID,
            initial_split_keys=INITIAL_SPLIT_KEYS,
            column_families=EMPTY_COLUMN_FAMILIES,
            task_id="id",
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        mock_hook.return_value.get_instance.return_value = None
        with pytest.raises(AirflowException) as ctx:
            op.execute(None)
        err = ctx.value
        assert str(err) == f"Dependency: instance '{INSTANCE_ID}' does not exist in project '{PROJECT_ID}'."
        mock_hook.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )

    @mock.patch("airflow.providers.google.cloud.operators.bigtable.BigtableHook")
    def test_creating_table_that_exists(self, mock_hook):
        op = BigtableCreateTableOperator(
            project_id=PROJECT_ID,
            instance_id=INSTANCE_ID,
            table_id=TABLE_ID,
            initial_split_keys=INITIAL_SPLIT_KEYS,
            column_families=EMPTY_COLUMN_FAMILIES,
            task_id="id",
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )

        mock_hook.return_value.get_column_families_for_table.return_value = EMPTY_COLUMN_FAMILIES
        instance = mock_hook.return_value.get_instance.return_value = mock.Mock(Instance)
        mock_hook.return_value.create_table.side_effect = mock.Mock(
            side_effect=google.api_core.exceptions.AlreadyExists("Table already exists.")
        )
        op.execute(None)

        mock_hook.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        mock_hook.return_value.create_table.assert_called_once_with(
            instance=instance,
            table_id=TABLE_ID,
            initial_split_keys=INITIAL_SPLIT_KEYS,
            column_families=EMPTY_COLUMN_FAMILIES,
        )

    @mock.patch("airflow.providers.google.cloud.operators.bigtable.BigtableHook")
    def test_creating_table_that_exists_empty_project_id(self, mock_hook):
        op = BigtableCreateTableOperator(
            instance_id=INSTANCE_ID,
            table_id=TABLE_ID,
            initial_split_keys=INITIAL_SPLIT_KEYS,
            column_families=EMPTY_COLUMN_FAMILIES,
            task_id="id",
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )

        mock_hook.return_value.get_column_families_for_table.return_value = EMPTY_COLUMN_FAMILIES
        instance = mock_hook.return_value.get_instance.return_value = mock.Mock(Instance)
        mock_hook.return_value.create_table.side_effect = mock.Mock(
            side_effect=google.api_core.exceptions.AlreadyExists("Table already exists.")
        )
        op.execute(None)

        mock_hook.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        mock_hook.return_value.create_table.assert_called_once_with(
            instance=instance,
            table_id=TABLE_ID,
            initial_split_keys=INITIAL_SPLIT_KEYS,
            column_families=EMPTY_COLUMN_FAMILIES,
        )

    @mock.patch("airflow.providers.google.cloud.operators.bigtable.BigtableHook")
    def test_creating_table_that_exists_with_different_column_families_ids_in_the_table(self, mock_hook):
        op = BigtableCreateTableOperator(
            project_id=PROJECT_ID,
            instance_id=INSTANCE_ID,
            table_id=TABLE_ID,
            initial_split_keys=INITIAL_SPLIT_KEYS,
            column_families=EMPTY_COLUMN_FAMILIES,
            task_id="id",
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )

        mock_hook.return_value.get_column_families_for_table.return_value = {"existing_family": None}
        mock_hook.return_value.create_table.side_effect = mock.Mock(
            side_effect=google.api_core.exceptions.AlreadyExists("Table already exists.")
        )

        with pytest.raises(AirflowException) as ctx:
            op.execute(None)
        err = ctx.value
        assert str(err) == f"Table '{TABLE_ID}' already exists with different Column Families."
        mock_hook.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )

    @mock.patch("airflow.providers.google.cloud.operators.bigtable.BigtableHook")
    def test_creating_table_that_exists_with_different_column_families_gc_rule_in__table(self, mock_hook):
        op = BigtableCreateTableOperator(
            project_id=PROJECT_ID,
            instance_id=INSTANCE_ID,
            table_id=TABLE_ID,
            initial_split_keys=INITIAL_SPLIT_KEYS,
            column_families={"cf-id": MaxVersionsGCRule(1)},
            task_id="id",
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )

        cf_mock = mock.Mock()
        cf_mock.gc_rule = mock.Mock(return_value=MaxVersionsGCRule(2))

        mock_hook.return_value.get_column_families_for_table.return_value = {"cf-id": cf_mock}
        mock_hook.return_value.create_table.side_effect = mock.Mock(
            side_effect=google.api_core.exceptions.AlreadyExists("Table already exists.")
        )

        with pytest.raises(AirflowException) as ctx:
            op.execute(None)
        err = ctx.value
        assert str(err) == f"Table '{TABLE_ID}' already exists with different Column Families."
        mock_hook.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import copy
import logging
import os
import pathlib
import re
from collections import defaultdict
from collections.abc import Iterable
from typing import TYPE_CHECKING, Any

from google.cloud.dataproc_v1 import Batch, RuntimeConfig

from airflow.providers.common.compat.openlineage.facet import (
    ColumnLineageDatasetFacet,
    DatasetFacet,
    DocumentationDatasetFacet,
    Fields,
    Identifier,
    InputField,
    SchemaDatasetFacet,
    SchemaDatasetFacetFields,
    SymlinksDatasetFacet,
)
from airflow.providers.common.compat.openlineage.utils.spark import (
    inject_parent_job_information_into_spark_properties,
    inject_transport_information_into_spark_properties,
)
from airflow.providers.google.cloud.hooks.gcs import _parse_gcs_url

if TYPE_CHECKING:
    from google.cloud.bigquery.table import Table

    from airflow.providers.common.compat.openlineage.facet import Dataset
    from airflow.providers.common.compat.sdk import Context


log = logging.getLogger(__name__)

BIGQUERY_NAMESPACE = "bigquery"
BIGQUERY_URI = "bigquery"
WILDCARD = "*"


def merge_column_lineage_facets(facets: list[ColumnLineageDatasetFacet]) -> ColumnLineageDatasetFacet:
    """
    Merge multiple column lineage facets into a single consolidated facet.

    Specifically, it aggregates input fields and transformations for each field across all provided facets.

    Args:
        facets: Column Lineage Facets to be merged.

    Returns:
        A new Column Lineage Facet containing all fields, their respective input fields and transformations.

    Notes:
        - Input fields are uniquely identified by their `(namespace, name, field)` tuple.
        - If multiple facets contain the same field with the same input field, those input
          fields are merged without duplication.
        - Transformations associated with input fields are also merged. If transformations
          are not supported by the version of the `InputField` class, they will be omitted.
        - Transformation merging relies on a composite key of the field name and input field
          tuple to track and consolidate transformations.

    Examples:
        Case 1: Two facets with the same input field
        ```
        >>> facet1 = ColumnLineageDatasetFacet(
        ...     fields={"columnA": Fields(inputFields=[InputField("namespace1", "dataset1", "field1")])}
        ... )
        >>> facet2 = ColumnLineageDatasetFacet(
        ...     fields={"columnA": Fields(inputFields=[InputField("namespace1", "dataset1", "field1")])}
        ... )
        >>> merged = merge_column_lineage_facets([facet1, facet2])
        >>> merged.fields["columnA"].inputFields
        [InputField("namespace1", "dataset1", "field1")]
        ```

        Case 2: Two facets with different transformations for the same input field
        ```
        >>> facet1 = ColumnLineageDatasetFacet(
        ...     fields={
        ...         "columnA": Fields(
        ...             inputFields=[InputField("namespace1", "dataset1", "field1", transformations=["t1"])]
        ...         )
        ...     }
        ... )
        >>> facet2 = ColumnLineageDatasetFacet(
        ...     fields={
        ...         "columnA": Fields(
        ...             inputFields=[InputField("namespace1", "dataset1", "field1", transformations=["t2"])]
        ...         )
        ...     }
        ... )
        >>> merged = merge_column_lineage_facets([facet1, facet2])
        >>> merged.fields["columnA"].inputFields[0].transformations
        ["t1", "t2"]
        ```
    """
    # Dictionary to collect all unique input fields for each field name
    fields_sources: dict[str, set[tuple[str, str, str]]] = defaultdict(set)
    # Dictionary to aggregate transformations for each input field
    transformations: dict[str, list] = defaultdict(list)

    for facet in facets:
        for field_name, single_field in facet.fields.items():
            for input_field in single_field.inputFields:
                input_key_fields = (input_field.namespace, input_field.name, input_field.field)
                fields_sources[field_name].add(input_key_fields)

                if single_transformations := getattr(input_field, "transformations", []):
                    transformation_key = "".join((field_name, *input_key_fields))
                    transformations[transformation_key].extend(single_transformations)

    # Check if the `InputField` class supports the `transformations` attribute (since OL client 1.17.1)
    input_field_allows_transformation_info = True
    try:
        InputField(namespace="a", name="b", field="c", transformations=[])
    except TypeError:
        input_field_allows_transformation_info = False

    return ColumnLineageDatasetFacet(
        fields={
            field_name: Fields(
                inputFields=[
                    InputField(
                        namespace,
                        name,
                        column,
                        transformations.get("".join((field_name, namespace, name, column)), []),
                    )
                    if input_field_allows_transformation_info
                    else InputField(namespace, name, column)
                    for namespace, name, column in sorted(input_fields)
                ],
                transformationType="",  # Legacy transformation information
                transformationDescription="",  # Legacy transformation information
            )
            for field_name, input_fields in fields_sources.items()
        }
    )


def extract_ds_name_from_gcs_path(path: str) -> str:
    """
    Extract and process the dataset name from a given path.

    Args:
        path: The path to process e.g. of a gcs file.

    Returns:
        The processed dataset name.

    Examples:
        >>> extract_ds_name_from_gcs_path("/dir/file.*")
        'dir'
        >>> extract_ds_name_from_gcs_path("/dir/pre_")
        'dir'
        >>> extract_ds_name_from_gcs_path("/dir/file.txt")
        'dir/file.txt'
        >>> extract_ds_name_from_gcs_path("/dir/file.")
        'dir'
        >>> extract_ds_name_from_gcs_path("/dir/")
        'dir'
        >>> extract_ds_name_from_gcs_path("")
        '/'
        >>> extract_ds_name_from_gcs_path("/")
        '/'
        >>> extract_ds_name_from_gcs_path(".")
        '/'
    """
    if WILDCARD in path:
        path = path.split(WILDCARD, maxsplit=1)[0]

    # We want to end up with parent directory if the path:
    # - does not refer to a file (no dot in the last segment)
    #     and does not explicitly end with a slash, it is treated as a prefix and removed.
    #     Example: "/dir/pre_" -> "/dir/"
    # - contains a dot at the end, then it is treated as a prefix (created after removing the wildcard).
    #     Example: "/dir/file." (was "/dir/file.*" with wildcard) -> "/dir/"
    last_path_segment = os.path.basename(path).rstrip(".")
    if "." not in last_path_segment and not path.endswith("/"):
        path = pathlib.Path(path).parent.as_posix()

    # Normalize the path:
    # - Remove trailing slashes.
    # - Remove leading slashes.
    # - Handle edge cases for empty paths or single-dot paths.
    path = path.rstrip("/")
    path = path.lstrip("/")
    if path in ("", "."):
        path = "/"

    return path


def get_facets_from_bq_table(table: Table) -> dict[str, DatasetFacet]:
    """Get facets from BigQuery table object."""
    return get_facets_from_bq_table_for_given_fields(table, selected_fields=None)


def get_facets_from_bq_table_for_given_fields(
    table: Table, selected_fields: list[str] | None
) -> dict[str, DatasetFacet]:
    """
    Get facets from BigQuery table object for selected fields only.

    If selected_fields is None, include all fields.
    """
    facets: dict[str, DatasetFacet] = {}
    selected_fields_set = set(selected_fields) if selected_fields else None

    if table.schema:
        facets["schema"] = SchemaDatasetFacet(
            fields=[
                SchemaDatasetFacetFields(
                    name=schema_field.name, type=schema_field.field_type, description=schema_field.description
                )
                for schema_field in table.schema
                if selected_fields_set is None or schema_field.name in selected_fields_set
            ]
        )
    if table.description:
        facets["documentation"] = DocumentationDatasetFacet(description=table.description)

    if table.external_data_configuration:
        symlinks = get_namespace_name_from_source_uris(table.external_data_configuration.source_uris)
        facets["symlink"] = SymlinksDatasetFacet(
            identifiers=[
                Identifier(
                    namespace=namespace, name=name, type="file" if namespace.startswith("gs://") else "table"
                )
                for namespace, name in sorted(symlinks)
            ]
        )
    return facets


def get_namespace_name_from_source_uris(source_uris: Iterable[str]) -> set[tuple[str, str]]:
    result = set()
    for uri in source_uris:
        if uri.startswith("gs://"):
            bucket, blob = _parse_gcs_url(uri)
            result.add((f"gs://{bucket}", extract_ds_name_from_gcs_path(blob)))
        elif uri.startswith("https://googleapis.com/bigtable"):
            regex = r"https://googleapis.com/bigtable/projects/([^/]+)/instances/([^/]+)(?:/appProfiles/([^/]+))?/tables/([^/]+)"
            match = re.match(regex, uri)
            if match:
                project_id, instance_id, table_id = match.groups()[0], match.groups()[1], match.groups()[3]
                result.add((f"bigtable://{project_id}/{instance_id}", table_id))
    return result


def get_identity_column_lineage_facet(
    dest_field_names: Iterable[str],
    input_datasets: Iterable[Dataset],
) -> dict[str, DatasetFacet]:
    """
    Get column lineage facet for identity transformations.

    This function generates a simple column lineage facet, where each destination column
    consists of source columns of the same name from all input datasets that have that column.
    The lineage assumes there are no transformations applied, meaning the columns retain their
    identity between the source and destination datasets.

    Args:
        dest_field_names: A list of destination column names for which lineage should be determined.
        input_datasets: A list of input datasets with schema facets.

    Returns:
        A dictionary containing a single key, `columnLineage`, mapped to a `ColumnLineageDatasetFacet`.
         If no column lineage can be determined, an empty dictionary is returned - see Notes below.

    Notes:
        - If any input dataset lacks a schema facet, the function immediately returns an empty dictionary.
        - If any field in the source dataset's schema is not present in the destination table,
          the function returns an empty dictionary. The destination table can contain extra fields, but all
          source columns should be present in the destination table.
        - If none of the destination columns can be matched to input dataset columns, an empty
          dictionary is returned.
        - Extra columns in the destination table that do not exist in the input datasets are ignored and
          skipped in the lineage facet, as they cannot be traced back to a source column.
        - The function assumes there are no transformations applied, meaning the columns retain their
          identity between the source and destination datasets.
    """
    fields_sources: dict[str, list[Dataset]] = {}
    for ds in input_datasets:
        if not ds.facets or "schema" not in ds.facets:
            return {}
        for schema_field in ds.facets["schema"].fields:  # type: ignore[attr-defined]
            if schema_field.name not in dest_field_names:
                return {}
            fields_sources[schema_field.name] = fields_sources.get(schema_field.name, []) + [ds]

    if not fields_sources:
        return {}

    column_lineage_facet = ColumnLineageDatasetFacet(
        fields={
            field_name: Fields(
                inputFields=[
                    InputField(namespace=dataset.namespace, name=dataset.name, field=field_name)
                    for dataset in source_datasets
                ],
                transformationType="IDENTITY",
                transformationDescription="identical",
            )
            for field_name, source_datasets in fields_sources.items()
        }
    )
    return {"columnLineage": column_lineage_facet}


def get_from_nullable_chain(source: Any, chain: list[str]) -> Any | None:
    """
    Get object from nested structure of objects, where it's not guaranteed that all keys in the nested structure exist.

    Intended to replace chain of `dict.get()` statements.

    Example usage:

    .. code-block:: python

        if (
            not job._properties.get("statistics")
            or not job._properties.get("statistics").get("query")
            or not job._properties.get("statistics").get("query").get("referencedTables")
        ):
            return None
        result = job._properties.get("statistics").get("query").get("referencedTables")

    becomes:

    .. code-block:: python

        result = get_from_nullable_chain(properties, ["statistics", "query", "queryPlan"])
        if not result:
            return None
    """
    # chain.pop modifies passed list, this can be unexpected
    chain = chain.copy()
    chain.reverse()
    try:
        while chain:
            while isinstance(source, list) and len(source) == 1:
                source = source[0]
            next_key = chain.pop()
            if isinstance(source, dict):
                source = source.get(next_key)
            else:
                source = getattr(source, next_key)
        return source
    except AttributeError:
        return None


def _is_openlineage_provider_accessible() -> bool:
    """
    Check if the OpenLineage provider is accessible.

    This function attempts to import the necessary OpenLineage modules and checks if the provider
    is enabled and the listener is available.

    Returns:
        bool: True if the OpenLineage provider is accessible, False otherwise.
    """
    try:
        from airflow.providers.openlineage.conf import is_disabled
        from airflow.providers.openlineage.plugins.listener import get_openlineage_listener
    except ImportError:
        log.debug("OpenLineage provider could not be imported.")
        return False

    if is_disabled():
        log.debug("OpenLineage provider is disabled.")
        return False

    if not get_openlineage_listener():
        log.debug("OpenLineage listener could not be found.")
        return False

    return True


def _extract_supported_job_type_from_dataproc_job(job: dict) -> str | None:
    """
    Extract job type from a Dataproc job definition.

    Args:
        job: The Dataproc job definition.

    Returns:
        The job type for which the automatic OL injection is supported, if found, otherwise None.
    """
    supported_job_types = ("sparkJob", "pysparkJob", "spark_job", "pyspark_job")
    return next((job_type for job_type in supported_job_types if job_type in job), None)


def _replace_dataproc_job_properties(job: dict, job_type: str, new_properties: dict) -> dict:
    """
    Replace the properties of a specific job type in a Dataproc job definition.

    Args:
        job: The original Dataproc job definition.
        job_type: The key representing the job type (e.g., "sparkJob").
        new_properties: The new properties to replace the existing ones.

    Returns:
        A modified copy of the job with updated properties.

    Raises:
        KeyError: If the job_type does not exist in the job or lacks a "properties" field.
    """
    if job_type not in job:
        raise KeyError(f"Job type '{job_type}' is missing in the job definition.")

    updated_job = job.copy()
    updated_job[job_type] = job[job_type].copy()
    updated_job[job_type]["properties"] = new_properties

    return updated_job


def inject_openlineage_properties_into_dataproc_job(
    job: dict, context: Context, inject_parent_job_info: bool, inject_transport_info: bool
) -> dict:
    """
    Inject OpenLineage properties into Spark job definition.

    This function does not remove existing configurations or modify the job definition in any way,
     except to add the required OpenLineage properties if they are not already present.

    The entire properties injection process will be skipped if any condition is met:
        - The OpenLineage provider is not accessible.
        - The job type is unsupported.
        - Both `inject_parent_job_info` and `inject_transport_info` are set to False.

    Additionally, specific information will not be injected if relevant OpenLineage properties already exist.

    Parent job information will not be injected if:
        - Any property prefixed with `spark.openlineage.parent` exists.
        - `inject_parent_job_info` is False.
    Transport information will not be injected if:
        - Any property prefixed with `spark.openlineage.transport` exists.
        - `inject_transport_info` is False.

    Args:
        job: The original Dataproc job definition.
        context: The Airflow context in which the job is running.
        inject_parent_job_info: Flag indicating whether to inject parent job information.
        inject_transport_info: Flag indicating whether to inject transport information.

    Returns:
        The modified job definition with OpenLineage properties injected, if applicable.
    """
    if not inject_parent_job_info and not inject_transport_info:
        log.debug("Automatic injection of OpenLineage information is disabled.")
        return job

    if not _is_openlineage_provider_accessible():
        log.warning(
            "Could not access OpenLineage provider for automatic OpenLineage "
            "properties injection. No action will be performed."
        )
        return job

    if (job_type := _extract_supported_job_type_from_dataproc_job(job)) is None:
        log.warning(
            "Could not find a supported Dataproc job type for automatic OpenLineage "
            "properties injection. No action will be performed.",
        )
        return job

    properties = job[job_type].get("properties", {})

    if inject_parent_job_info:
        log.debug("Injecting OpenLineage parent job information into Spark properties.")
        properties = inject_parent_job_information_into_spark_properties(
            properties=properties, context=context
        )

    if inject_transport_info:
        log.debug("Injecting OpenLineage transport information into Spark properties.")
        properties = inject_transport_information_into_spark_properties(
            properties=properties, context=context
        )

    job_with_ol_config = _replace_dataproc_job_properties(
        job=job, job_type=job_type, new_properties=properties
    )
    return job_with_ol_config


def _is_dataproc_batch_of_supported_type(batch: dict | Batch) -> bool:
    """
    Check if a Dataproc batch is of a supported type for Openlineage automatic injection.

    This function determines if the given batch is of a supported type
    by checking for specific job type attributes or keys in the batch.

    Args:
        batch: The Dataproc batch to check.

    Returns:
        True if the batch is of a supported type (`spark_batch` or
        `pyspark_batch`), otherwise False.
    """
    supported_job_types = ("spark_batch", "pyspark_batch")
    if isinstance(batch, Batch):
        if any(getattr(batch, job_type) for job_type in supported_job_types):
            return True
        return False

    # For dictionary-based batch
    if any(job_type in batch for job_type in supported_job_types):
        return True
    return False


def _extract_dataproc_batch_properties(batch: dict | Batch) -> dict:
    """
    Extract Dataproc batch properties from a Batch object or dictionary.

    This function retrieves the `properties` from the `runtime_config` of a
    Dataproc `Batch` object or a dictionary representation of a batch.

    Args:
        batch: The Dataproc batch to extract properties from.

    Returns:
        Extracted `properties` if found, otherwise an empty dictionary.
    """
    if isinstance(batch, Batch):
        return dict(batch.runtime_config.properties)

    # For dictionary-based batch
    run_time_config = batch.get("runtime_config", {})
    if isinstance(run_time_config, RuntimeConfig):
        return dict(run_time_config.properties)
    return run_time_config.get("properties", {})


def _replace_dataproc_batch_properties(batch: dict | Batch, new_properties: dict) -> dict | Batch:
    """
    Replace the properties of a Dataproc batch.

    Args:
        batch: The original Dataproc batch definition.
        new_properties: The new properties to replace the existing ones.

    Returns:
        A modified copy of the Dataproc batch definition with updated properties.
    """
    batch = copy.deepcopy(batch)
    if isinstance(batch, Batch):
        if not batch.runtime_config:
            batch.runtime_config = RuntimeConfig(properties=new_properties)
        elif isinstance(batch.runtime_config, dict):
            batch.runtime_config["properties"] = new_properties
        else:
            batch.runtime_config.properties = new_properties
        return batch

    # For dictionary-based batch
    run_time_config = batch.get("runtime_config")
    if not run_time_config:
        batch["runtime_config"] = {"properties": new_properties}
    elif isinstance(run_time_config, dict):
        run_time_config["properties"] = new_properties
    else:
        run_time_config.properties = new_properties
    return batch


def inject_openlineage_properties_into_dataproc_batch(
    batch: dict | Batch, context: Context, inject_parent_job_info: bool, inject_transport_info: bool
) -> dict | Batch:
    """
    Inject OpenLineage properties into Dataproc batch definition.

    This function does not remove existing configurations or modify the batch definition in any way,
     except to add the required OpenLineage properties if they are not already present.

    The entire properties injection process will be skipped if any condition is met:
        - The OpenLineage provider is not accessible.
        - The batch type is unsupported.
        - Both `inject_parent_job_info` and `inject_transport_info` are set to False.

    Additionally, specific information will not be injected if relevant OpenLineage properties already exist.

    Parent job information will not be injected if:
        - Any property prefixed with `spark.openlineage.parent` exists.
        - `inject_parent_job_info` is False.
    Transport information will not be injected if:
        - Any property prefixed with `spark.openlineage.transport` exists.
        - `inject_transport_info` is False.

    Args:
        batch: The original Dataproc batch definition.
        context: The Airflow context in which the job is running.
        inject_parent_job_info: Flag indicating whether to inject parent job information.
        inject_transport_info: Flag indicating whether to inject transport information.

    Returns:
        The modified batch definition with OpenLineage properties injected, if applicable.
    """
    if not inject_parent_job_info and not inject_transport_info:
        log.debug("Automatic injection of OpenLineage information is disabled.")
        return batch

    if not _is_openlineage_provider_accessible():
        log.warning(
            "Could not access OpenLineage provider for automatic OpenLineage "
            "properties injection. No action will be performed."
        )
        return batch

    if not _is_dataproc_batch_of_supported_type(batch):
        log.warning(
            "Could not find a supported Dataproc batch type for automatic OpenLineage "
            "properties injection. No action will be performed.",
        )
        return batch

    properties = _extract_dataproc_batch_properties(batch)

    if inject_parent_job_info:
        log.debug("Injecting OpenLineage parent job information into Spark properties.")
        properties = inject_parent_job_information_into_spark_properties(
            properties=properties, context=context
        )

    if inject_transport_info:
        log.debug("Injecting OpenLineage transport information into Spark properties.")
        properties = inject_transport_information_into_spark_properties(
            properties=properties, context=context
        )

    batch_with_ol_config = _replace_dataproc_batch_properties(batch=batch, new_properties=properties)
    return batch_with_ol_config


def inject_openlineage_properties_into_dataproc_workflow_template(
    template: dict, context: Context, inject_parent_job_info: bool, inject_transport_info: bool
) -> dict:
    """
    Inject OpenLineage properties into all Spark jobs within Workflow Template.

    This function does not remove existing configurations or modify the jobs definition in any way,
     except to add the required OpenLineage properties if they are not already present.

    The entire properties injection process for each job will be skipped if any condition is met:
        - The OpenLineage provider is not accessible.
        - The job type is unsupported.
        - Both `inject_parent_job_info` and `inject_transport_info` are set to False.

    Additionally, specific information will not be injected if relevant OpenLineage properties already exist.

    Parent job information will not be injected if:
        - Any property prefixed with `spark.openlineage.parent` exists.
        - `inject_parent_job_info` is False.
    Transport information will not be injected if:
        - Any property prefixed with `spark.openlineage.transport` exists.
        - `inject_transport_info` is False.

    Args:
        template: The original Dataproc Workflow Template definition.
        context: The Airflow context in which the job is running.
        inject_parent_job_info: Flag indicating whether to inject parent job information.
        inject_transport_info: Flag indicating whether to inject transport information.

    Returns:
        The modified Workflow Template definition with OpenLineage properties injected, if applicable.
    """
    if not inject_parent_job_info and not inject_transport_info:
        log.debug("Automatic injection of OpenLineage information is disabled.")
        return template

    if not _is_openlineage_provider_accessible():
        log.warning(
            "Could not access OpenLineage provider for automatic OpenLineage "
            "properties injection. No action will be performed."
        )
        return template

    final_jobs = []
    for single_job_definition in template["jobs"]:
        step_id = single_job_definition["step_id"]
        log.debug("Injecting OpenLineage properties into Workflow step: `%s`", step_id)

        if (job_type := _extract_supported_job_type_from_dataproc_job(single_job_definition)) is None:
            log.debug(
                "Could not find a supported Dataproc job type for automatic OpenLineage "
                "properties injection. No action will be performed.",
            )
            final_jobs.append(single_job_definition)
            continue

        properties = single_job_definition[job_type].get("properties", {})

        if inject_parent_job_info:
            log.debug("Injecting OpenLineage parent job information into Spark properties.")
            properties = inject_parent_job_information_into_spark_properties(
                properties=properties, context=context
            )

        if inject_transport_info:
            log.debug("Injecting OpenLineage transport information into Spark properties.")
            properties = inject_transport_information_into_spark_properties(
                properties=properties, context=context
            )

        job_with_ol_config = _replace_dataproc_job_properties(
            job=single_job_definition, job_type=job_type, new_properties=properties
        )
        final_jobs.append(job_with_ol_config)

    template["jobs"] = final_jobs
    return template
